from cs_kit import CoreTestFunctions
from cs_config import functions


OK_ADJUSTMENT = {
    "policy": {
        "STD": [
            {"MARS": "single", "year": 2019, "value": 0},
            {"MARS": "mjoint", "year": 2019, "value": 1},
            {"MARS": "mjoint", "year": 2022, "value": 10},
        ],
        "parameter_indexing_CPI_offset": [{"year": 2019, "value": -0.001}],
        "EITC_c": [{"EIC": "0kids", "year": 2019, "value": 1000.0}],
    },
    "behavior": {"inc": [{"value": -0.1}]},
}


BAD_ADJUSTMENT = {
    "policy": {
        "STD": [
            {"MARS": "single", "year": 2019, "value": -10},
            {"MARS": "mjoint", "year": 2019, "value": 1},
            {"MARS": "mjoint", "year": 2022, "value": 10},
        ],
        "parameter_indexing_CPI_offset": [{"year": 2019, "value": -0.001}],
        "ACTC_c": [{"year": 2019, "value": 2000.0}],
    },
    "behavior": {"sub": [{"value": -0.1}]},
}

EMPTY_ADJUSTMENT = {"policy": {}, "behavior": {}}


CHECKBOX_ADJUSTMENT = {
    "policy": {
        "STD": [
            {"MARS": "single", "year": 2019, "value": 10},
            {"MARS": "mjoint", "year": 2019, "value": 1},
            {"MARS": "mjoint", "year": 2022, "value": 10},
        ],
        "STD_checkbox": [{"value": False}],
    },
    "behavior": {},
}


def test_start_year_with_data_source():
    """
    Test interaction between PUF and CPS data sources and the start year.
    """
    # Commenting out PUF for now since not on CS web app
    # data = functions.get_inputs({"data_source": "PUF"})
    # assert (
    #     data["meta_parameters"]["year"]["validators"]["choice"]["choices"][0]
    #     == 2013
    # )
    data = functions.get_inputs({"data_source": "CPS"})
    assert (
        data["meta_parameters"]["year"]["validators"]["choice"]["choices"][0]
        == 2014
    )

    ew = {
        "policy": {"errors": {}, "warnings": {}},
        "behavior": {"errors": {}, "warnings": {}},
    }
    res = functions.validate_inputs(
        {"data_source": "CPS", "year": 2013},
        {"policy": {}, "behavior": {}},
        ew,
    )
    assert res["errors_warnings"]["policy"]["errors"].get("year")


# def test_start_year_with_data_source():
#     """
#     Test interaction between PUF and CPS data sources and the start year.
#     """
#     data = functions.get_inputs({"data_source": "PUF"})
#     assert data["meta_parameters"]["year"]["validators"]["choice"]["choices"][0] == 2013
#     data = functions.get_inputs({"data_source": "CPS"})
#     assert data["meta_parameters"]["year"]["validators"]["choice"]["choices"][0] == 2014

#     ew = {
#         "policy": {"errors": {}, "warnings": {}},
#         "behavior": {"errors": {}, "warnings": {}},
#     }
#     res = functions.validate_inputs(
#         {"data_source": "CPS", "year": 2013}, {"policy": {}, "behavior": {}}, ew
#     )
#     assert res["errors_warnings"]["policy"]["errors"].get("year")


class TestFunctions1(CoreTestFunctions):
    get_version = functions.get_version
    get_inputs = functions.get_inputs
    validate_inputs = functions.validate_inputs
    run_model = functions.run_model
    ok_adjustment = OK_ADJUSTMENT
    bad_adjustment = BAD_ADJUSTMENT


class TestFunctions2(CoreTestFunctions):
    get_version = functions.get_version
    get_inputs = functions.get_inputs
    validate_inputs = functions.validate_inputs
    run_model = functions.run_model
    ok_adjustment = CHECKBOX_ADJUSTMENT
    bad_adjustment = BAD_ADJUSTMENT


class TestFunctions3(CoreTestFunctions):
    get_version = functions.get_version
    get_inputs = functions.get_inputs
    validate_inputs = functions.validate_inputs
    run_model = functions.run_model
    ok_adjustment = EMPTY_ADJUSTMENT
    bad_adjustment = BAD_ADJUSTMENT



import taxbrain
import pytest


def test_distribution_plot(tb_static):
    tb_static.run()
    fig = taxbrain.distribution_plot(tb_static, 2019)


def test_differences_plot(tb_static):
    tb_static.run()
    fig = taxbrain.differences_plot(tb_static, "combined")
    with pytest.raises(AssertionError):
        taxbrain.differences_plot(tb_static, "wages")


def test_volcano_plot(tb_static):
    tb_static.run()
    fig = taxbrain.volcano_plot(tb_static, 2019)
    with pytest.raises(ValueError):
        taxbrain.volcano_plot(tb_static, 2019, min_y=-10000)
    fig = taxbrain.volcano_plot(tb_static, 2019, min_y=-1000, log_scale=False)
    # testing using RGB tuples for the colors
    fig = taxbrain.volcano_plot(
        tb_static,
        2019,
        increase_color=(0.1, 0.2, 0.5),
        decrease_color=(0.2, 0.2, 0.5),
    )


def test_lorenz_curve(tb_static):
    tb_static.run()
    fig = taxbrain.lorenz_curve(tb_static, 2019)


def test_revenue_plot(tb_static):
    tb_static.run()
    fig = taxbrain.revenue_plot(tb_static)
    with pytest.raises(ValueError):
        taxbrain.revenue_plot(tb_static, tax_vars=["income", "combined"])
    with pytest.raises(AssertionError):
        taxbrain.revenue_plot(tb_static, tax_vars=[])



"""
Functions to create the tables displayed in COMP
"""

from .constants import AGG_ROW_NAMES
from taxbrain import TaxBrain


def summary_aggregate(res, tb):
    """
    Function to produce aggregate revenue tables
    """
    if not isinstance(res, dict):
        raise TypeError(
            f"'res' is of type {type(res)}. Must be dictionary object."
        )
    if not isinstance(tb, TaxBrain):
        raise TypeError(f"'tb' is of type {type(tb)}. Must be TaxBrain object")
    # tax totals for baseline
    tax_vars = ["iitax", "payrolltax", "combined", "benefit_cost_total"]
    aggr_base = tb.multi_var_table(tax_vars, "base")
    aggr_base.index = AGG_ROW_NAMES
    # tax totals for reform
    aggr_reform = tb.multi_var_table(tax_vars, "reform")
    aggr_reform.index = AGG_ROW_NAMES
    # tax difference
    aggr_d = aggr_reform - aggr_base
    # add to dictionary
    res["aggr_d"] = (aggr_d / 1e9).round(3)
    res["aggr_1"] = (aggr_base / 1e9).round(3)
    res["aggr_2"] = (aggr_reform / 1e9).round(3)
    del aggr_base, aggr_reform, aggr_d
    return res


def summary_dist_xbin(res, tb, year):
    """
    res is dictionary of summary-results DataFrames.
    df1 contains results variables for baseline policy.
    df2 contains results variables for reform policy.
    returns augmented dictionary of summary-results DataFrames.
    """
    if not isinstance(res, dict):
        raise TypeError(
            f"'res' is of type {type(res)}. Must be dictionary object."
        )
    if not isinstance(tb, TaxBrain):
        raise TypeError(f"'tb' is of type {type(tb)}. Must be TaxBrain object")
    if not isinstance(year, int):
        raise TypeError(f"'year' is of type {type(year)}. Must be an integer")
    # create distribution tables grouped by xbin
    # baseline distribution table
    res["dist1_xbin"] = tb.distribution_table(
        year, "standard_income_bins", "expanded_income", "base"
    )
    # reform distribution table
    # ensure income is grouped on the same measure
    expanded_income_baseline = tb.base_data[year]["expanded_income"]
    tb.reform_data[year]["expanded_income_baseline"] = expanded_income_baseline
    res["dist2_xbin"] = tb.distribution_table(
        year, "standard_income_bins", "expanded_income_baseline", "reform"
    )
    del tb.reform_data[year]["expanded_income_baseline"]
    return res


def summary_diff_xbin(res, tb, year):
    """
    res is dictionary of summary-results DataFrames.
    df1 contains results variables for baseline policy.
    df2 contains results variables for reform policy.
    returns augmented dictionary of summary-results DataFrames.
    """
    if not isinstance(res, dict):
        raise TypeError(
            f"'res' is of type {type(res)}. Must be dictionary object."
        )
    if not isinstance(tb, TaxBrain):
        raise TypeError(f"'tb' is of type {type(tb)}. Must be TaxBrain object")
    if not isinstance(year, int):
        raise TypeError(f"'year' is of type {type(year)}. Must be an integer")
    # create difference tables grouped by xbin
    res["diff_itax_xbin"] = tb.differences_table(
        year, "standard_income_bins", "iitax"
    )
    res["diff_ptax_xbin"] = tb.differences_table(
        year, "standard_income_bins", "payrolltax"
    )
    res["diff_comb_xbin"] = tb.differences_table(
        year, "standard_income_bins", "combined"
    )
    return res


def summary_dist_xdec(res, tb, year):
    """
    res is dictionary of summary-results DataFrames.
    df1 contains results variables for baseline policy.
    df2 contains results variables for reform policy.
    returns augmented dictionary of summary-results DataFrames.
    """
    if not isinstance(res, dict):
        raise TypeError(
            f"'res' is of type {type(res)}. Must be dictionary object."
        )
    if not isinstance(tb, TaxBrain):
        raise TypeError(f"'tb' is of type {type(tb)}. Must be TaxBrain object")
    if not isinstance(year, int):
        raise TypeError(f"'year' is of type {type(year)}. Must be an integer")
    # create distribution tables grouped by xdec
    res["dist1_xdec"] = tb.distribution_table(
        year, "weighted_deciles", "expanded_income", "base"
    )
    # reform distribution table
    # ensure income is grouped on the same measure
    expanded_income_baseline = tb.base_data[year]["expanded_income"]
    tb.reform_data[year]["expanded_income_baseline"] = expanded_income_baseline
    res["dist2_xdec"] = tb.distribution_table(
        year, "weighted_deciles", "expanded_income_baseline", "reform"
    )
    del tb.reform_data[year]["expanded_income_baseline"]
    return res


def summary_diff_xdec(res, tb, year):
    """
    res is dictionary of summary-results DataFrames.
    df1 contains results variables for baseline policy.
    df2 contains results variables for reform policy.
    returns augmented dictionary of summary-results DataFrames.
    """
    if not isinstance(res, dict):
        raise TypeError(
            f"'res' is of type {type(res)}. Must be dictionary object."
        )
    if not isinstance(tb, TaxBrain):
        raise TypeError(f"'tb' is of type {type(tb)}. Must be TaxBrain object")
    if not isinstance(year, int):
        raise TypeError(f"'year' is of type {type(year)}. Must be an integer")
    # create difference tables grouped by xdec
    res["diff_itax_xdec"] = tb.differences_table(
        year, "weighted_deciles", "iitax"
    )
    res["diff_ptax_xdec"] = tb.differences_table(
        year, "weighted_deciles", "payrolltax"
    )
    res["diff_comb_xdec"] = tb.differences_table(
        year, "weighted_deciles", "combined"
    )
    return res



# imports
import copy


# Default parameters for the CorporateIncidence class
CI_PARAMS = {
    "Incidence": {  # long-run incidence of corporate tax
        "Labor share": 0.5,
        "Shareholder share": 0.4,
        "All capital share": 0.1,
    },
    "Long run years": 10,  # number of years to reach long-run incidence
}

VAR_NAME_MAP = {
    "Dividends": "e00650",
    "Capital gains": "e01100",
    "Self-employment income": "e00900",
    "Taxable interest": "e00300",
    "Tax-exempt interest": "e00300",
}

SHORT_RUN_SHARES = {
    "Labor share": 0.0,
    "Shareholder share": 1.0,
    "All capital share": 0.0,
}


# TODO: think about if corp revenue intended to be entered in billions or what
def distribute(calc, corp_revenue, year, start_year, param_updates=None):
    """
    Function that distributes the corporate income tax incidence across
    individual income tax payers.
    """
    if param_updates:
        CI_PARAMS.update(param_updates)
    # Make a copy so as not to modify calculator object
    calc_corp = copy.deepcopy(calc)
    # Find index of year working on
    i = year - start_year
    # take revenue total for the relevant year
    if isinstance(corp_revenue, dict):
        corp_revenue = corp_revenue["year"]
    else:
        corp_revenue = corp_revenue[i]
    # adjust parameters for the transition to the long run
    shares = {}
    for k, v in CI_PARAMS["Incidence"].items():
        if CI_PARAMS["Long run years"] == 0:
            shares[k] = v
        else:
            increment = (v - SHORT_RUN_SHARES[k]) / CI_PARAMS["Long run years"]
            # use min() so don't "overshoot" the long run share
            shares[k] = (
                SHORT_RUN_SHARES[k]
                + min(i, CI_PARAMS["Long run years"]) * increment
            )

    # update income in the calc object
    calc_corp = _update_income(calc_corp, corp_revenue, shares)

    return calc_corp


def _update_income(calc, revenue, shares):
    """
    Implement income changes to account for corporate tax
    incidence on household filers
    """
    # Find aggregate change in wages
    agg_wage = shares["Labor share"] * revenue
    # Find aggregate change in dividend/capital gains
    agg_shareholder = shares["Shareholder share"] * revenue
    # Find aggregate change in other capital income
    agg_other_capital = shares["All capital share"] * revenue
    # With aggregates, compute pct change in each
    wage_income_vars = ["e00200"]
    shareholder_income_vars = ["p22250", "p23250", "e00600"]
    other_capital_income_vars = [
        "e00300",
        "e00400",
        "e01100",
        "e01200",
        "e02000",
    ]  # do we try to capture just some of e02000?
    denom = 0
    for v in wage_income_vars:
        denom += (calc.array(v) * calc.array("s006")).sum()
    pct_change_wage = agg_wage / denom
    denom = 0
    for v in shareholder_income_vars:
        denom += (calc.array(v) * calc.array("s006")).sum()
    pct_change_shareholder = agg_shareholder / denom
    denom = 0
    for v in other_capital_income_vars:
        denom += (calc.array(v) * calc.array("s006")).sum()
    pct_change_other_capital = agg_other_capital / denom

    # we will apply these percentage changes to the relevant income sources
    # need to scale other wages so their sum equals total
    for v in wage_income_vars + ["e00200p", "e00200s"]:
        delta = calc.array(v) * pct_change_wage
        calc.incarray(v, delta)
    for v in shareholder_income_vars:
        delta = calc.array(v) * pct_change_shareholder
        calc.incarray(v, delta)
    for v in other_capital_income_vars + ["e00650"]:
        # scale qualified dividends as well, though they are included
        # in e00600
        delta = calc.array(v) * pct_change_other_capital
        calc.incarray(v, delta)

    return calc



import shutil
from pathlib import Path
from taxbrain import report


def test_report(tb_static):
    """
    Ensure that all report files are created
    """
    outdir = "testreform"
    name = "Test Report"
    report(tb_static, name=name, outdir=outdir)
    dir_path = Path(outdir)
    assert dir_path.exists()
    assert Path(dir_path, "Test-Report.md").exists()
    assert Path(dir_path, "Test-Report.pdf").exists()
    diff_png = Path(dir_path, "difference_graph.png")
    assert diff_png.exists()
    dist_png = Path(dir_path, "dist_graph.png")
    assert dist_png.exists()
    shutil.rmtree(dir_path)
    # test clean report
    _content = report(tb_static, name=name, outdir=outdir, clean=True)
    assert not dir_path.exists()



"""
Functions for creating the TaxBrain COMP outputs
"""

from bokeh.models import (
    ColumnDataSource,
    Toggle,
    CustomJS,
    NumeralTickFormatter,
    HoverTool,
)
from bokeh.models import TabPanel, Tabs, Div
from bokeh.embed import json_item
from bokeh.layouts import layout
from bokeh.plotting import figure


def aggregate_plot(tb):
    """
    Function for creating a bokeh plot that shows aggregate tax liabilities for
    each year the TaxBrain instance was run
    Parameters
    ----------
    tb: An instance of the TaxBrain object
    Returns
    -------
    Bokeh figure
    """
    # Pull aggregate data by year and transpose it for plotting
    varlist = ["iitax", "payrolltax", "combined"]
    base_data = tb.multi_var_table(varlist, "base").transpose()
    base_data["calc"] = "Base"
    reform_data = tb.multi_var_table(varlist, "reform").transpose()
    reform_data["calc"] = "Reform"
    base_cds = ColumnDataSource(base_data)
    reform_cds = ColumnDataSource(reform_data)
    num_ticks = len(base_data)
    del base_data, reform_data

    fig = figure(
        title="Aggregate Tax Liability by Year",
        width=700,
        height=500,
        tools="save",
    )
    ii_base = fig.line(
        x="index",
        y="iitax",
        line_width=4,
        line_color="#12719e",
        legend_label="Income Tax - Base",
        source=base_cds,
    )
    ii_reform = fig.line(
        x="index",
        y="iitax",
        line_width=4,
        line_color="#73bfe2",
        legend_label="Income Tax - Reform",
        source=reform_cds,
    )
    proll_base = fig.line(
        x="index",
        y="payrolltax",
        line_width=4,
        line_color="#408941",
        legend_label="Payroll Tax - Base",
        source=base_cds,
    )
    proll_reform = fig.line(
        x="index",
        y="payrolltax",
        line_width=4,
        line_color="#98cf90",
        legend_label="Payroll Tax - Reform",
        source=reform_cds,
    )
    comb_base = fig.line(
        x="index",
        y="combined",
        line_width=4,
        line_color="#a4201d",
        legend_label="Combined - Base",
        source=base_cds,
    )
    comb_reform = fig.line(
        x="index",
        y="combined",
        line_width=4,
        line_color="#e9807d",
        legend_label="Combined - Reform",
        source=reform_cds,
    )

    # format figure
    fig.legend.location = "top_left"
    fig.yaxis.formatter = NumeralTickFormatter(format="$0.00a")
    fig.yaxis.axis_label = "Aggregate Tax Liability"
    fig.xaxis.minor_tick_line_color = None
    fig.xaxis[0].ticker.desired_num_ticks = num_ticks

    # Add hover tool
    tool_str = """
        <p><b>@calc - {}</b></p>
        <p>${}</p>
    """
    ii_hover = HoverTool(
        tooltips=tool_str.format("Individual Income Tax", "@iitax{0,0}"),
        renderers=[ii_base, ii_reform],
    )
    proll_hover = HoverTool(
        tooltips=tool_str.format("Payroll Tax", "@payrolltax{0,0}"),
        renderers=[proll_base, proll_reform],
    )
    combined_hover = HoverTool(
        tooltips=tool_str.format("Combined Tax", "@combined{0,0}"),
        renderers=[comb_base, comb_reform],
    )
    fig.add_tools(ii_hover, proll_hover, combined_hover)

    # toggle which lines are shown
    plot_js = """
    object1.visible = toggle.active
    object2.visible = toggle.active
    object3.visible = toggle.active
    """
    base_callback = CustomJS(code=plot_js, args={})
    base_toggle = Toggle(label="Base", button_type="primary", active=True)
    base_callback.args = {
        "toggle": base_toggle,
        "object1": ii_base,
        "object2": proll_base,
        "object3": comb_base,
    }
    base_toggle.js_on_change("active", base_callback)

    reform_callback = CustomJS(code=plot_js, args={})
    reform_toggle = Toggle(label="Reform", button_type="primary", active=True)
    reform_callback.args = {
        "toggle": reform_toggle,
        "object1": ii_reform,
        "object2": proll_reform,
        "object3": comb_reform,
    }
    fig_layout = layout([fig], [base_toggle, reform_toggle])
    reform_toggle.js_on_change("active", reform_callback)

    # Components needed to embed the figure
    data = json_item(fig_layout)
    outputs = {
        "media_type": "bokeh",
        "title": "",
        "data": data,
    }

    return outputs


def create_layout(data, start_year, end_year):
    """
    Function for creating a bokeh layout with all of the data tables
    """

    agg_data = data["aggr_outputs"]
    # create aggregate table
    clt_title = f"<h3>{agg_data['current']['title']}</h3>"
    current_law_table = Div(
        text=clt_title + agg_data["current"]["renderable"], width=1000
    )
    rt_title = f"<h3>{agg_data['reform']['title']}</h3>"
    reform_table = Div(
        text=rt_title + agg_data["reform"]["renderable"], width=1000
    )
    ct_title = f"<h3>{agg_data['change']['title']}</h3>"
    change_table = Div(
        text=ct_title + agg_data["change"]["renderable"], width=1000
    )

    current_tab = TabPanel(child=current_law_table, title="Current Law")
    reform_tab = TabPanel(child=reform_table, title="Reform")
    change_tab = TabPanel(child=change_table, title="Change")
    agg_tabs = Tabs(tabs=[current_tab, reform_tab, change_tab])

    key_map = {
        "current": "Current",
        "reform": "Reform",
        "ind_income": "Income Tax",
        "payroll": "Payroll Tax",
        "combined": "Combined Tax",
        "dist": "Distribution Table",
        "diff": "Differences Table",
    }

    tbl_data = data["tbl_outputs"]
    yr_panels = []
    # loop through each year (start - end year)
    for yr in range(start_year, end_year + 1):
        # loop through each table type: dist, idff
        tbl_panels = []
        for tbl_type, content in tbl_data.items():
            # loop through sub tables: current, reform for dist
            # ind_income, payroll, combined for diff
            content_panels = []
            for key, value in content.items():
                # loop through each grouping: bins, deciles
                grp_panels = []
                for grp, grp_data in value.items():
                    _data = grp_data[yr]
                    # create a data table for this tab
                    title = f"<h3>{_data['title']}</h3>"
                    note = (
                        "<p><i>All monetary totals are in billions. "
                        "All counts are in millions. "
                        "Averages and shares are as shown.</i></p>"
                    )
                    tbl = Div(
                        text=title + note + _data["renderable"], width=1000
                    )
                    grp_panel = TabPanel(child=tbl, title=grp.title())
                    grp_panels.append(grp_panel)
                grp_tab = Tabs(tabs=grp_panels)
                # panel for the sub tables
                content_panel = TabPanel(child=grp_tab, title=key_map[key])
                content_panels.append(content_panel)
            content_tab = Tabs(tabs=content_panels)
            # panel for the table types
            tbl_panel = TabPanel(child=content_tab, title=key_map[tbl_type])
            tbl_panels.append(tbl_panel)
        type_tab = Tabs(tabs=tbl_panels)
        # panel for the year
        yr_panel = TabPanel(child=type_tab, title=str(yr))
        yr_panels.append(yr_panel)

    yr_tabs = Tabs(tabs=yr_panels)

    agg_layout = layout(children=[agg_tabs])
    table_layout = layout(children=[yr_tabs])
    agg_data = json_item(agg_layout)
    table_data = json_item(table_layout)

    # return a dictionary of outputs ready for COMP
    agg_outputs = {
        "media_type": "bokeh",
        "title": "Aggregate Results",
        "data": agg_data,
    }
    table_outputs = {
        "media_type": "bokeh",
        "title": "Tables",
        "data": table_data,
    }

    # return js, div, cdn_js, cdn_css, widget_js, widget_css
    return agg_outputs, table_outputs



# imports
import numpy as np
import pandas as pd
import taxcalc as tc
import copy
import pytest
from taxbrain import corporate_incidence


# Tests of corporate_incidence functions


def test_update_income_shares():
    """
    Test of the _update_income() function
    """
    # create a calculator object
    rec = tc.Records.cps_constructor()
    pol = tc.Policy()
    calc1 = tc.Calculator(policy=pol, records=rec)

    # set a corp revenue amount
    revenue = 100_000_000_00.0

    # set shares
    shares = {
        "Labor share": 0.2,
        "Shareholder share": 0.7,
        "All capital share": 0.1,
    }
    # use update income function
    test_calc = corporate_incidence._update_income(
        copy.deepcopy(calc1), revenue, shares
    )

    wage_income_vars = ["e00200"]
    shareholder_income_vars = ["p22250", "p23250", "e00600"]
    other_capital_income_vars = [
        "e00300",
        "e00400",
        "e01100",
        "e01200",
        "e02000",
    ]

    # Check that shares match expected
    sum_test = 0
    sum_base = 0
    for v in wage_income_vars:
        sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
        sum_base += (calc1.array(v) * calc1.array("s006")).sum()
    pct = (sum_test - sum_base) / revenue
    assert np.allclose(pct, shares["Labor share"])
    sum_test = 0
    sum_base = 0
    for v in shareholder_income_vars:
        sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
        sum_base += (calc1.array(v) * calc1.array("s006")).sum()
    pct = (sum_test - sum_base) / revenue
    assert np.allclose(pct, shares["Shareholder share"])
    sum_test = 0
    sum_base = 0
    for v in other_capital_income_vars:
        sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
        sum_base += (calc1.array(v) * calc1.array("s006")).sum()
    pct = (sum_test - sum_base) / revenue
    assert np.allclose(pct, shares["All capital share"])


def test_update_income_totals():
    """
    Test of the _update_income() function
    """
    # create a calculator object
    rec = tc.Records.cps_constructor()
    pol = tc.Policy()
    calc1 = tc.Calculator(policy=pol, records=rec)

    # set a corp revenue amount
    revenue = 100_000_000_00.0

    # set shares
    shares = {
        "Labor share": 0.2,
        "Shareholder share": 0.7,
        "All capital share": 0.1,
    }
    # use update income function
    test_calc = corporate_incidence._update_income(
        copy.deepcopy(calc1), revenue, shares
    )

    income_vars = [
        "e00200",
        "p22250",
        "p23250",
        "e00600",
        "e00300",
        "e00400",
        "e01100",
        "e01200",
        "e02000",
    ]
    # check that in total, amount of income change equals revenue
    sum_test = 0
    sum_base = 0
    for v in income_vars:
        sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
        sum_base += (calc1.array(v) * calc1.array("s006")).sum()
    diff = sum_test - sum_base
    assert np.allclose(revenue, diff)


expected_shares5 = {
    "Labor share": {
        2025: 0.0,
        2026: 0.04,
        2027: 0.08,
        2028: 0.12,
        2029: 0.16,
        2030: 0.2,
        2031: 0.2,
        2032: 0.2,
        2033: 0.2,
    },
    "Shareholder share": {
        2025: 1.0,
        2026: 0.94,
        2027: 0.88,
        2028: 0.82,
        2029: 0.76,
        2030: 0.7,
        2031: 0.7,
        2032: 0.7,
        2033: 0.7,
    },
    "All capital share": {
        2025: 0.0,
        2026: 0.02,
        2027: 0.04,
        2028: 0.06,
        2029: 0.08,
        2030: 0.1,
        2031: 0.1,
        2032: 0.1,
        2033: 0.1,
    },
}
expected_shares0 = {
    "Labor share": {
        2025: 0.2,
        2026: 0.2,
        2027: 0.2,
        2028: 0.2,
    },
    "Shareholder share": {
        2025: 0.7,
        2026: 0.7,
        2027: 0.7,
        2028: 0.7,
    },
    "All capital share": {
        2025: 0.1,
        2026: 0.1,
        2027: 0.1,
        2028: 0.1,
    },
}
expected_shares1 = {
    "Labor share": {
        2025: 0.0,
        2026: 0.2,
        2027: 0.2,
        2028: 0.2,
    },
    "Shareholder share": {
        2025: 1.0,
        2026: 0.7,
        2027: 0.7,
        2028: 0.7,
    },
    "All capital share": {
        2025: 0.0,
        2026: 0.1,
        2027: 0.1,
        2028: 0.1,
    },
}


@pytest.mark.parametrize(
    "transition_years,end_year,expected_shares",
    [
        (5, 2033, expected_shares5),
        (0, 2028, expected_shares0),
        (1, 2028, expected_shares1),
    ],
    ids=["5 years", "immediate", "1 year"],
)
def test_share_transition(transition_years, end_year, expected_shares):
    """
    Test the transition of share over time in the
    corp_incidence_distribute function.
    """
    # create a calculator object
    rec = tc.Records.cps_constructor()
    pol = tc.Policy()
    calc1 = tc.Calculator(policy=pol, records=rec)

    # set a corp revenue amount
    revenue = [100_000_000_00.0] * (2034 - 2025)

    # Define different income sources
    # TODO: should make these constants in corp_incidence.py
    wage_income_vars = ["e00200"]
    shareholder_income_vars = ["p22250", "p23250", "e00600"]
    other_capital_income_vars = [
        "e00300",
        "e00400",
        "e01100",
        "e01200",
        "e02000",
    ]
    # Long run shares
    param_updates = {
        "Incidence": {
            "Labor share": 0.2,
            "Shareholder share": 0.7,
            "All capital share": 0.1,
        },
        "Long run years": transition_years,
    }

    # Now check shares match expectation for each year
    for year in range(2025, end_year + 1):
        print("Checking for year ", year)
        calc1.advance_to_year(year)
        test_calc = corporate_incidence.distribute(
            copy.deepcopy(calc1), revenue, year, 2025, param_updates
        )

        # Check that shares match expected
        sum_test = 0
        sum_base = 0
        for v in wage_income_vars:
            sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
            sum_base += (calc1.array(v) * calc1.array("s006")).sum()
        pct1 = (sum_test - sum_base) / revenue
        assert np.allclose(pct1, expected_shares["Labor share"][year])
        sum_test = 0
        sum_base = 0
        for v in shareholder_income_vars:
            sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
            sum_base += (calc1.array(v) * calc1.array("s006")).sum()
        pct2 = (sum_test - sum_base) / revenue
        assert np.allclose(pct2, expected_shares["Shareholder share"][year])
        sum_test = 0
        sum_base = 0
        for v in other_capital_income_vars:
            sum_test += (test_calc.array(v) * test_calc.array("s006")).sum()
            sum_base += (calc1.array(v) * calc1.array("s006")).sum()
        pct3 = (sum_test - sum_base) / revenue
        assert np.allclose(pct3, expected_shares["All capital share"][year])

        # also check that shares always sum to 1
        assert np.allclose(1.0, pct1 + pct2 + pct3)


# Validation check
# Try to use same parameters and recreate TPC
# (https://www.taxpolicycenter.org/sites/default/files/alfresco/publication-pdfs/412651-How-TPC-Distributes-the-Corporate-Income-Tax.PDF)
# Table 8, with percent of burden by income quantile
def test_validation():
    """
    Test the transition of share over time in the
    corp_incidence_distribute function.
    """
    # create a calculator object
    rec = tc.Records.cps_constructor()
    pol = tc.Policy()
    calc1 = tc.Calculator(policy=pol, records=rec)

    # set a corp revenue amount
    revenue = [100_000_000_00.0] * (2034 - 2025)

    # Define different income sources
    # TODO: should make these constants in corp_incidence.py
    wage_income_vars = ["e00200"]
    shareholder_income_vars = ["p22250", "p23250", "e00600", "e00650"]
    other_capital_income_vars = [
        "e00300",
        "e00400",
        "e01100",
        "e01200",
        "e02000",
    ]
    # Long run shares
    param_updates = {
        "Incidence": {
            "Labor share": 0.0,
            "Shareholder share": 1.0,
            "All capital share": 0.0,
        },
        "Long run years": 0,
    }

    # expected shares if full burden on shareholders:
    expected = pd.DataFrame(
        data=[
            3.12755650e-05,
            0.00000000e00,
            9.41971262e-04,
            3.66821040e-03,
            3.90232598e-03,
            8.00940431e-03,
            1.28149782e-02,
            2.03518571e-02,
            2.64449904e-02,
            5.06202883e-02,
            7.93136449e-02,
            7.93901054e-01,
            1.00000000e00,
            7.38510850e-02,
            1.25578069e-01,
            5.94471900e-01,
        ],
        index=[
            "0-10n",
            "0-10z",
            "0-10p",
            "10-20",
            "20-30",
            "30-40",
            "40-50",
            "50-60",
            "60-70",
            "70-80",
            "80-90",
            "90-100",
            "ALL",
            "90-95",
            "95-99",
            "Top 1%",
        ],
    )

    test_calc = corporate_incidence.distribute(
        copy.deepcopy(calc1), revenue, 2025, 2025, param_updates
    )
    calc1.advance_to_year(2025)
    test_calc.advance_to_year(2025)
    print("Calc current year = ", calc1.current_year)
    print("Test Calc current year = ", test_calc.current_year)
    calc1.calc_all()
    test_calc.calc_all()
    dt1, dt2 = calc1.distribution_tables(test_calc, "weighted_deciles")
    print("Distribution table: ")
    total_income_change = (
        dt2.loc["ALL", "expanded_income"] - dt1.loc["ALL", "expanded_income"]
    )
    print("total income change = ", total_income_change)
    pct_change_income = (
        dt2["expanded_income"] - dt1["expanded_income"]
    ) / total_income_change
    print(np.array(pct_change_income.values))
    print(expected[0].values)

    assert np.allclose(
        np.array(pct_change_income.values), expected[0].values, atol=1e-5
    )



# constants used by the new compconfig
import paramtools
from marshmallow import fields, Schema
from datetime import datetime
from taxbrain import TaxBrain


POLICY_SCHEMA = {
    "labels": {
        "year": {
            "type": "int",
            "validators": {
                "choice": {
                    "choices": [
                        yr
                        for yr in range(
                            TaxBrain.FIRST_BUDGET_YEAR,
                            TaxBrain.LAST_BUDGET_YEAR + 1,
                        )
                    ]
                }
            },
        },
        "MARS": {
            "type": "str",
            "validators": {
                "choice": {
                    "choices": [
                        "single",
                        "mjoint",
                        "mseparate",
                        "headhh",
                        "widow",
                    ]
                }
            },
        },
        "idedtype": {
            "type": "str",
            "validators": {
                "choice": {
                    "choices": [
                        "med",
                        "sltx",
                        "retx",
                        "cas",
                        "misc",
                        "int",
                        "char",
                    ]
                }
            },
        },
        "EIC": {
            "type": "str",
            "validators": {
                "choice": {"choices": ["0kids", "1kid", "2kids", "3+kids"]}
            },
        },
        "data_source": {
            "type": "str",
            "validators": {"choice": {"choices": ["PUF", "CPS", "other"]}},
        },
    },
    "additional_members": {
        "section_1": {"type": "str"},
        "section_2": {"type": "str"},
        "start_year": {"type": "int"},
        "checkbox": {"type": "bool"},
    },
}

AGG_ROW_NAMES = [
    "ind_tax",
    "payroll_tax",
    "combined_tax",
    "benefit_cost_total",
]

RESULTS_TABLE_TITLES = {
    "diff_comb_xbin": (
        "Combined Payroll and Individual Income Tax: Difference"
        " between Base and User plans by expanded income bin"
    ),
    "diff_comb_xdec": (
        "Combined Payroll and Individual Income Tax: Difference"
        " between Base and User plans by expanded income "
        "decile"
    ),
    "diff_itax_xbin": (
        "Individual Income Tax: Difference between Base and "
        "User plans by expanded income bin"
    ),
    "diff_itax_xdec": (
        "Individual Income Tax: Difference between Base and "
        "User plans by expanded income decile"
    ),
    "diff_ptax_xbin": (
        "Payroll Tax: Difference between Base and User plans "
        "by expanded income bin"
    ),
    "diff_ptax_xdec": (
        "Payroll Tax: Difference between Base and User plans "
        "by expanded income decile"
    ),
    "dist1_xbin": "Base plan tax vars, weighted total by expanded income bin",
    "dist1_xdec": (
        "Base plan tax vars, weighted total by expanded income " "decile"
    ),
    "dist2_xbin": "User plan tax vars, weighted total by expanded income bin",
    "dist2_xdec": (
        "User plan tax vars, weighted total by expanded income " "decile"
    ),
    "aggr_1": "Total Liabilities Baseline by Calendar Year (Billions)",
    "aggr_d": "Total Liabilities Change by Calendar Year (Billions)",
    "aggr_2": "Total Liabilities Reform by Calendar Year (Billions)",
}

RESULTS_TABLE_TAGS = {
    # diff tables
    "diff_comb_xbin": {
        "table_type": "diff",
        "tax_type": "combined",
        "grouping": "bins",
    },
    "diff_comb_xdec": {
        "table_type": "diff",
        "tax_type": "combined",
        "grouping": "deciles",
    },
    "diff_itax_xbin": {
        "table_type": "diff",
        "tax_type": "ind_income",
        "grouping": "bins",
    },
    "diff_itax_xdec": {
        "table_type": "diff",
        "tax_type": "ind_income",
        "grouping": "deciles",
    },
    "diff_ptax_xbin": {
        "table_type": "diff",
        "tax_type": "payroll",
        "grouping": "bins",
    },
    "diff_ptax_xdec": {
        "table_type": "diff",
        "tax_type": "payroll",
        "grouping": "deciles",
    },
    # dist tables
    "dist1_xbin": {"table_type": "dist", "law": "current", "grouping": "bins"},
    "dist1_xdec": {
        "table_type": "dist",
        "law": "current",
        "grouping": "deciles",
    },
    "dist2_xbin": {"table_type": "dist", "law": "reform", "grouping": "bins"},
    "dist2_xdec": {
        "table_type": "dist",
        "law": "reform",
        "grouping": "deciles",
    },
    # aggr tables
    "aggr_1": {"law": "current"},
    "aggr_d": {"law": "change"},
    "aggr_2": {"law": "reform"},
    # gdp elaticity model table
    "gdp_effect": {"default": "gdp_elast"},
}
RESULTS_TOTAL_ROW_KEY_LABELS = {
    "ind_tax": "Individual Income Tax Liability",
    "payroll_tax": "Payroll Tax Liability",
    "combined_tax": (
        "Combined Payroll and Individual Income Tax " + "Liability"
    ),
    "benefit_cost_total": "Total Benefits Spending",
}

MONEY_VARS = {
    "AGI",
    "Standard Deduction",
    "Itemized Deduction",
    "Personal Exemption",
    "Taxable Income",
    "Regular Tax",
    "AMTI",
    "Tax before Credits",
    "Non-refundable Credits",
    "Other Taxes",
    "Refundable Credits",
    "Individual Income Tax Liabilities",
    "Payroll Tax Liablities",
    "Combined Payroll and Individual Income Tax Liabilities",
    "Universal Basic Income",
    "Total Cost of Benefits",
    "Consumption Value of Benefits",
    "Expanded Income",
    "After-Tax Expanded Income",
    "Average Tax Change",
    "Total Tax Difference",
}


class MetaParameters(paramtools.Parameters):
    array_first = True
    defaults = {
        "year": {
            "title": "Start Year",
            "description": "Year for parameters.",
            "type": "int",
            "value": min(datetime.now().year, TaxBrain.LAST_BUDGET_YEAR),
            "validators": {
                "when": {
                    "param": "data_source",
                    "is": "CPS",
                    "then": {
                        "range": {
                            "min": 2014,
                            "max": TaxBrain.LAST_BUDGET_YEAR,
                        }
                    },
                    "otherwise": {
                        "range": {
                            "min": 2013,
                            "max": TaxBrain.LAST_BUDGET_YEAR,
                        }
                    },
                }
            },
        },
        "data_source": {
            "title": "Data Source",
            "description": "CPS is currently the only supported data source",
            "type": "str",
            "value": "CPS",
            "validators": {"choice": {"choices": ["CPS"]}},
        },
        "use_full_sample": {
            "title": "Use Full Sample",
            "description": "Use entire data set or a 2% sample.",
            "type": "bool",
            "value": True,
            "validators": {"choice": {"choices": [True, False]}},
        },
    }

    def dump(self, *args, **kwargs):
        """
        This method extends the default ParamTools dump method by swapping the
        when validator for a choice validator. This is required because C/S does
        not yet implement the when validator.
        """
        data = super().dump(*args, **kwargs)
        if self.data_source == "CPS":
            data["year"]["validators"] = {
                "choice": {
                    "choices": list(range(2014, TaxBrain.LAST_BUDGET_YEAR))
                }
            }
        else:
            data["year"]["validators"] = {
                "choice": {
                    "choices": list(range(2013, TaxBrain.LAST_BUDGET_YEAR))
                }
            }
        return data



import os
import pytest
import pandas as pd
from taxbrain import TaxBrain


CUR_PATH = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture(scope="session")
def reform_json_str():
    reform = """
        {
            "policy": {
                "SS_thd1": {"2019": [50000, 100000, 50000, 50000, 50000]},
                "SS_thd2": {"2019": [50000, 100000, 50000, 50000, 50000]},
                "SS_Earnings_thd": {"2019": 400000},
                "FICA_ss_trt_employee": {"2020": 0.0625,
                                "2021": 0.063,
                                "2022": 0.0635,
                                "2023": 0.064,
                                "2024": 0.0645,
                                "2025": 0.065,
                                "2026": 0.0605,
                                "2027": 0.061,
                                "2028": 0.0615},
                "STD-indexed": {"2019": false}
            }
        }
    """
    return reform


@pytest.fixture(scope="session")
def assump_json_str():
    assump = """
        {
            "consumption": {"BEN_housing_value": {"2019": 0.7}},
            "growdiff_baseline": {"ABOOK": {"2019": 0.01}},
            "growdiff_response": {"ACGNS": {"2019": 0.01}}
        }
    """
    return assump


@pytest.fixture(
    scope="session",
)
def tb_static(reform_json_str):
    return TaxBrain(2018, 2019, microdata="CPS", reform=reform_json_str)


@pytest.fixture(scope="session")
def tb_dynamic(reform_json_str):
    return TaxBrain(
        2018,
        2019,
        microdata="CPS",
        reform=reform_json_str,
        behavior={"sub": 0.25},
    )


@pytest.fixture(scope="session")
def empty_mods():
    return {
        "consumption": {},
        "growdiff_response": {},
        "policy": {},
        "growdiff_baseline": {},
        "behavior": {},
    }


@pytest.fixture(scope="session")
def sample_input():
    params = {
        "params": {
            "policy": {"_FICA_ss_trt_employee": {"2019": [0.075]}},
            "behavior": {"sub": {"2019": [0.1]}},
        },
        "jsonstrs": "",
        "errors_warnings": {
            "policy": {"errors": {}, "warnings": {}},
            "behavior": {"errors": {}, "warnings": {}},
        },
        "start_year": 2019,
        "data_source": "PUF",
        "use_full_sample": False,
    }
    return params



"""
Functions used to help tax-brain configure to COMP
"""

import os
import inspect
import time
import copy
import hashlib
import gzip
import copy
from pathlib import Path
import warnings

import pandas as pd
import numpy as np
from collections import defaultdict
from taxbrain.report_utils import convert_params
from taxcalc import (
    Policy,
    DIFF_TABLE_COLUMNS,
    DIFF_TABLE_LABELS,
    DIST_TABLE_COLUMNS,
    DIST_TABLE_LABELS,
    add_income_table_row_variable,
    add_quantile_table_row_variable,
    STANDARD_INCOME_BINS,
)
from operator import itemgetter
from .constants import (
    POLICY_SCHEMA,
    RESULTS_TABLE_TAGS,
    RESULTS_TABLE_TITLES,
    RESULTS_TOTAL_ROW_KEY_LABELS,
    MONEY_VARS,
)
from .tables import (
    summary_aggregate,
    summary_diff_xbin,
    summary_diff_xdec,
    summary_dist_xbin,
    summary_dist_xdec,
)

try:
    from s3fs import S3FileSystem
except ImportError as ie:
    S3FileSystem = None

TCPATH = inspect.getfile(Policy)
TCDIR = os.path.dirname(TCPATH)


AWS_ACCESS_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID", None)
AWS_SECRET_ACCESS_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY", None)

PUF_S3_FILE_LOCATION = os.environ.get(
    "PUF_S3_LOCATION", "s3://ospc-data-files/puf.20210720.csv.gz"
)

TMD_S3_FILE_LOCATION = os.environ.get(
    "TMD_S3_LOCATION", "s3://ospc-data-files/tmd.20210720.csv.gz"
)


def random_seed(user_mods, year):
    """
    Compute random seed based on specified user_mods, which is a
    dictionary returned by Calculator.read_json_parameter_files().
    """

    def random_seed_from_subdict(subdict):
        """
        Compute random seed from one user_mods subdictionary.
        """
        assert isinstance(subdict, dict)
        all_vals = []
        for year in sorted(subdict.keys()):
            all_vals.append(str(year))
            params = subdict[year]
            for param in sorted(params.keys()):
                try:
                    tple = tuple(params[param])
                except TypeError:
                    # params[param] is not an iterable value; make it so
                    tple = tuple((params[param],))
                all_vals.append(str((param, tple)))
        txt = "".join(all_vals).encode("utf-8")
        hsh = hashlib.sha512(txt)
        seed = int(hsh.hexdigest(), 16)
        return seed % np.iinfo(np.uint32).max

    # start of random_seed function
    # modify the user mods to work in the random_seed_from_subdict function
    # TODO: Change all of this to work with new adjustments
    user_mods_copy = copy.deepcopy(user_mods)
    beh_mods_dict = {year: {}}
    for param, value in user_mods_copy["behavior"].items():
        beh_mods_dict[year][param] = [value]
    user_mods_copy["behavior"] = beh_mods_dict
    ans = 0
    for subdict_name in user_mods_copy:
        subdict = user_mods_copy[subdict_name]
        if subdict_name == "policy":
            subdict = convert_params(subdict)
        ans += random_seed_from_subdict(subdict)
    return ans % np.iinfo(np.uint32).max


NUM_TO_FUZZ = 3  # when using dropq algorithm on puf.csv results


def fuzzed(df1, df2, reform_affected, table_row_type):
    """
    Create fuzzed df2 dataframe and corresponding unfuzzed df1 dataframe.

    Parameters
    ----------
    df1: Pandas DataFrame
        contains results variables for the baseline policy, which are not
        changed by this function

    df2: Pandas DataFrame
        contains results variables for the reform policy, which are not
        changed by this function

    reform_affected: boolean numpy array (not changed by this function)
        True for filing units with a reform-induced combined tax difference;
        otherwise False

    table_row_type: string
        valid values are 'aggr', 'xbin', and 'xdec'

    Returns
    -------
    df1, df2: Pandas DataFrames
        where copied df2 is fuzzed to maintain data privacy and
        where copied df1 has same filing unit order as has the fuzzed df2
    """
    assert table_row_type in ("aggr", "xbin", "xdec")
    assert len(df1.index) == len(df2.index)
    assert reform_affected.size == len(df1.index)
    df1 = copy.deepcopy(df1)
    df2 = copy.deepcopy(df2)
    # add copy of reform_affected to df2
    df2["reform_affected"] = copy.deepcopy(reform_affected)
    # construct table rows, for which filing units in each row must be fuzzed
    if table_row_type == "xbin":
        df1 = add_income_table_row_variable(
            df1, "expanded_income", STANDARD_INCOME_BINS
        )
        df2["expanded_income_baseline"] = df1["expanded_income"]
        df2 = add_income_table_row_variable(
            df2, "expanded_income_baseline", STANDARD_INCOME_BINS
        )
        del df2["expanded_income_baseline"]
    elif table_row_type == "xdec":
        df1 = add_quantile_table_row_variable(
            df1, "expanded_income", 10, decile_details=True
        )
        df2["expanded_income_baseline"] = df1["expanded_income"]
        df2 = add_quantile_table_row_variable(
            df2, "expanded_income_baseline", 10, decile_details=True
        )
        del df2["expanded_income_baseline"]
    elif table_row_type == "aggr":
        df1["table_row"] = np.ones(reform_affected.shape, dtype=int)
        df2["table_row"] = df1["table_row"]
    gdf1 = df1.groupby("table_row", sort=False)
    gdf2 = df2.groupby("table_row", sort=False)
    del df1["table_row"]
    del df2["table_row"]
    # fuzz up to NUM_TO_FUZZ filing units randomly chosen in each group
    # (or table row), where fuzz means to replace the reform (2) results
    # with the baseline (1) results for each chosen filing unit
    pd.options.mode.chained_assignment = None
    group_list = list()
    for name, group2 in gdf2:
        group2 = copy.deepcopy(group2)
        indices = np.where(group2["reform_affected"])
        num = min(len(indices[0]), NUM_TO_FUZZ)
        if num > 0:
            choices = np.random.choice(indices[0], size=num, replace=False)
            group1 = gdf1.get_group(name)
            for idx in choices:
                group2.iloc[idx] = group1.iloc[idx]
        group_list.append(group2)
        del group2
    df2 = pd.concat(group_list)
    del df2["reform_affected"]
    pd.options.mode.chained_assignment = "warn"
    # reinstate index order of df1 and df2 and return
    df1.sort_index(inplace=True)
    df2.sort_index(inplace=True)
    return (df1, df2)


def nth_year_results(tb, year, user_mods, fuzz, return_html=True):
    """
    Function to process taxbrain results for a given year
    """
    start_time = time.time()
    dv1 = tb.base_data[year]
    dv2 = tb.reform_data[year]
    sres = {}
    if fuzz:
        # seed random number generator with a seed value based on user_mods
        # (reform-specific seed is used to choose whose results are fuzzed)
        seed = random_seed(user_mods, year)
        np.random.seed(seed)
        # make bool array marking which filing units are affected by the reform
        reform_affected = np.logical_not(
            np.isclose(dv1["combined"], dv2["combined"], atol=0.01, rtol=0.0)
        )
        agg1, agg2 = fuzzed(dv1, dv2, reform_affected, "aggr")
        sres = summary_aggregate(sres, tb)
        del agg1
        del agg2
        dv1b, dv2b = fuzzed(dv1, dv2, reform_affected, "xbin")
        sres = summary_dist_xbin(sres, tb, year)
        sres = summary_diff_xbin(sres, tb, year)
        del dv1b
        del dv2b
        dv1d, dv2d = fuzzed(dv1, dv2, reform_affected, "xdec")
        sres = summary_dist_xdec(sres, tb, year)
        sres = summary_diff_xdec(sres, tb, year)
        del dv1d
        del dv2d
        del reform_affected
    else:
        sres = summary_aggregate(sres, tb)
        sres = summary_dist_xbin(sres, tb, year)
        sres = summary_diff_xbin(sres, tb, year)
        sres = summary_dist_xdec(sres, tb, year)
        sres = summary_diff_xdec(sres, tb, year)

    # optionally return non-JSON-like results
    # it would be nice to allow the user to download the full CSV instead
    # of a CSV for each year
    # what if we allowed an aggregate format call?
    #  - presents project with all data proeduced in a run?

    if return_html:
        res = {}
        for id in sres:
            res[id] = [{"dimension": year, "raw": sres[id]}]
        elapsed_time = time.time() - start_time
        print("elapsed time for this run: {:.1f}".format(elapsed_time))
        return res
    else:
        elapsed_time = time.time() - start_time
        print("elapsed time for this run: {:.1f}".format(elapsed_time))
        return sres


def postprocess(data_to_process):
    """
    Receives results from run_nth_year_taxcalc_model over N years,
    formats the results, and combines the aggregate results
    """
    labels = {
        x: DIFF_TABLE_LABELS[i] for i, x in enumerate(DIFF_TABLE_COLUMNS)
    }
    labels.update(
        {x: DIST_TABLE_LABELS[i] for i, x in enumerate(DIST_TABLE_COLUMNS)}
    )

    # nested functions used below
    def label_columns(pdf):
        pdf.columns = [
            (labels[str(col)] if str(col) in labels else str(col))
            for col in pdf.columns
        ]
        return pdf

    def append_year(pdf, year):
        """
        append_year embedded function revises all column names in dframe
        """
        pdf.columns = ["{}_{}".format(col, year) for col in pdf.columns]
        return pdf

    def year_columns(pdf, year):
        pdf.columns = [str(year)]
        return pdf

    def arbitrary_defaultdict():
        """
        Return an arbitrary number of defaultdicts. This is used to store all
        of the distribution and differences tables
        """
        return defaultdict(arbitrary_defaultdict)

    formatted = {
        "tbl_outputs": arbitrary_defaultdict(),
        "aggr_outputs": defaultdict(dict),
    }
    downloadable = []
    year_getter = itemgetter("dimension")
    for id, pdfs in data_to_process.items():
        if id.startswith("aggr"):
            pdfs.sort(key=year_getter)
            tbl = pdfs[0]["raw"]
            tbl.index = pd.Index(
                RESULTS_TOTAL_ROW_KEY_LABELS[i] for i in tbl.index
            )
            # format table
            for col in tbl.columns:
                tbl[col] = tbl[col].apply("${:,.2f}".format)

            title = RESULTS_TABLE_TITLES[id]
            tags = RESULTS_TABLE_TAGS[id]
            formatted["aggr_outputs"][tags["law"]] = {
                "title": title,
                "renderable": pdf_to_clean_html(tbl),
            }
            # append a downloadable version of the results
            downloadable.append(
                {
                    "media_type": "CSV",
                    "title": title + ".csv",
                    "data": tbl.to_csv(),
                }
            )

        else:
            for i in pdfs:
                year = i["dimension"]
                tbl = label_columns(i["raw"])
                title = "{} ({})".format(RESULTS_TABLE_TITLES[id], year)
                # format table
                for col in tbl.columns:
                    if col in MONEY_VARS:
                        tbl[col] = tbl[col].apply("${:,.2f}".format)

                tags = RESULTS_TABLE_TAGS[id]
                tbl_type = tags["table_type"]
                group = tags["grouping"]
                if id.startswith("dist"):
                    law = tags["law"]
                    formatted["tbl_outputs"][tbl_type][law][group][year] = {
                        "title": title,
                        "renderable": pdf_to_clean_html(tbl),
                    }
                else:
                    tax = tags["tax_type"]
                    formatted["tbl_outputs"][tbl_type][tax][group][year] = {
                        "title": title,
                        "renderable": pdf_to_clean_html(tbl),
                    }

                # add downloadable information
                downloadable.append(
                    {
                        "media_type": "CSV",
                        "title": title + ".csv",
                        "data": tbl.to_csv(),
                    }
                )

    return formatted, downloadable


def pdf_to_clean_html(pdf):
    """Takes a PDF and returns an HTML table without any deprecated tags or
    irrelevant styling"""
    tb_replace = '<table class="table table-striped"'

    return (
        pdf.to_html()
        .replace("<table ", tb_replace)
        .replace(' border="1"', "")
        .replace('class="dataframe"', "")
    )


def retrieve_puf(
    puf_s3_file_location=PUF_S3_FILE_LOCATION,
    aws_access_key_id=AWS_ACCESS_KEY_ID,
    aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
):
    """
    Function for retrieving the PUF from the S3 bucket
    """
    s3_reader_installed = S3FileSystem is not None
    has_credentials = (
        aws_access_key_id is not None and aws_secret_access_key is not None
    )
    if puf_s3_file_location and has_credentials and s3_reader_installed:
        print("Reading puf from S3 bucket.", puf_s3_file_location)
        fs = S3FileSystem(
            key=AWS_ACCESS_KEY_ID,
            secret=AWS_SECRET_ACCESS_KEY,
        )
        with fs.open(puf_s3_file_location) as f:
            # Skips over header from top of file.
            puf_df = pd.read_csv(f)
        return puf_df
    elif Path("puf.csv.gz").exists():
        print("Reading puf from puf.csv.gz.")
        return pd.read_csv("puf.csv.gz", compression="gzip")
    elif Path("puf.csv").exists():
        print("Reading puf from puf.csv.")
        return pd.read_csv("puf.csv")
    else:
        warnings.warn(
            f"PUF file not available (puf_location={puf_s3_file_location}, "
            f"has_credentials={has_credentials}, "
            f"s3_reader_installed={s3_reader_installed})"
        )
        return None


def retrieve_tmd(
    tmd_s3_file_location=TMD_S3_FILE_LOCATION,
    aws_access_key_id=AWS_ACCESS_KEY_ID,
    aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
):
    """
    Function for retrieving the TMD from the S3 bucket
    """
    s3_reader_installed = S3FileSystem is not None
    has_credentials = (
        aws_access_key_id is not None and aws_secret_access_key is not None
    )
    if tmd_s3_file_location and has_credentials and s3_reader_installed:
        print("Reading tmd from S3 bucket.", tmd_s3_file_location)
        fs = S3FileSystem(
            key=AWS_ACCESS_KEY_ID,
            secret=AWS_SECRET_ACCESS_KEY,
        )
        with fs.open(tmd_s3_file_location) as f:
            # Skips over header from top of file.
            tmd_df = pd.read_csv(f)
        return tmd_df
    elif Path("tmd.csv.gz").exists():
        print("Reading tmd from tmd.csv.gz.")
        return pd.read_csv("tmd.csv.gz", compression="gzip")
    elif Path("tmd.csv").exists():
        print("Reading tmd from tmd.csv.")
        return pd.read_csv("tmd.csv")
    else:
        warnings.warn(
            f"TMD file not available (tmd_location={tmd_s3_file_location}, "
            f"has_credentials={has_credentials}, "
            f"s3_reader_installed={s3_reader_installed})"
        )
        return None



import shutil
from taxbrain import cli_core
from pathlib import Path


def test_cli():
    """
    Test core logic of the CLI
    """
    startyear = 2019
    endyear = 2020
    data = "CPS"
    reform = None
    behavior = None
    assump = None
    baseline = None
    outdir = ""
    name = "test_cli"
    make_report = False
    author = None
    cli_core(
        startyear,
        endyear,
        data,
        reform,
        behavior,
        assump,
        baseline,
        outdir,
        name,
        make_report,
        author,
    )
    outpath = Path(outdir, name)
    # assert that all folders and files have been created
    assert outpath.is_dir()
    assert Path(outpath, "aggregate_tax_liability.csv").exists()
    for year in range(startyear, endyear + 1):
        yearpath = Path(outpath, str(year))
        assert yearpath.is_dir()
        assert Path(yearpath, f"distribution_table_base_{year}.csv").exists()
        assert Path(yearpath, f"distribution_table_reform_{year}.csv").exists()
        assert Path(yearpath, f"differences_table_{year}.csv").exists()
    shutil.rmtree(outpath)



from taxbrain.taxbrain import *
from taxbrain.utils import *
from taxbrain.cli import *
from taxbrain.report import *
from taxbrain.report_utils import *

__version__ = "2.8.0"
__min_python3_version__ = 11
__max_python3_version__ = 13



"""
Helper functions for the various taxbrain modules
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.ticker as ticker
from collections import defaultdict
from typing import Union, Tuple

import taxcalc as tc
from .typing import ParamToolsAdjustment, TaxcalcReform, PlotColors


def weighted_sum(df, var, wt="s006"):
    """
    Return the weighted sum of specified variable

    Parameters
    ----------
    df: Pandas DataFrame
        data overwhich to compute weighted sum
    var: str
        variable name from df for which to computer weighted sum
    wt: str
        name of weight variable in df

    Returns
    -------
    float
        weighted sum
    """
    return (df[var] * df[wt]).sum()


def distribution_plot(
    tb,
    year: int,
    figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
    title: str = "default",
    include_text: bool = False,
):
    """
    Create a horizontal bar chart to display the distributional change in
    after tax income

    Parameters
    ----------
    tb: TaxBrain object
        TaxBrain object for analysis
    year: int
        year to report distribution for
    figsize: tuple
        representing the size of the figure (width, height) in inches
    title: str
        title for plot
    include_text: bool
        whether to include text for labels

    Returns
    -------
    fig: Matplotlib.pyplot figure object
        distribution plot
    """

    def find_percs(data, group):
        """
        Find the percentage of people in the data set that saw
        their income change by the given percentages
        """
        pop = data["s006"].sum()
        large_pos_chng = data["s006"][data["pct_change"] > 5].sum() / pop
        small_pos_chng = (
            data["s006"][
                (data["pct_change"] <= 5) & (data["pct_change"] > 1)
            ].sum()
            / pop
        )
        small_chng = (
            data["s006"][
                (data["pct_change"] <= 1) & (data["pct_change"] >= -1)
            ].sum()
            / pop
        )
        small_neg_change = (
            data["s006"][
                (data["pct_change"] < -1) & (data["pct_change"] > -5)
            ].sum()
            / pop
        )
        large_neg_change = data["s006"][data["pct_change"] < -5].sum() / pop

        return (
            large_pos_chng,
            small_pos_chng,
            small_chng,
            small_neg_change,
            large_neg_change,
        )

    # extract needed data from the TaxBrain object
    ati_data = pd.DataFrame(
        {
            "base": tb.base_data[year]["aftertax_income"],
            "reform": tb.reform_data[year]["aftertax_income"],
            "s006": tb.base_data[year]["s006"],
        }
    )
    ati_data["diff"] = ati_data["reform"] - ati_data["base"]
    ati_data["pct_change"] = (ati_data["diff"] / ati_data["base"]) * 100
    ati_data = ati_data.fillna(0.0)  # fill in NaNs for graphing
    # group tupules: (low income, high income, income group name)
    groups = [
        (-9e99, 9e99, "All"),
        (1e6, 9e99, "$1M or More"),
        (500000, 1e6, "$500K-1M"),
        (200000, 500000, "$200K-500K"),
        (100000, 200000, "$100K-200K"),
        (75000, 100000, "$75K-100K"),
        (50000, 75000, "$50K-75K"),
        (40000, 50000, "$40K-50K"),
        (30000, 40000, "$30K-40K"),
        (20000, 30000, "$20K-30K"),
        (10000, 20000, "$10K-20K"),
        (-9e99, 10000, "Less than $10K"),
    ]

    plot_data = defaultdict(list)
    # traverse list in reverse to get the axis of the plot in correct order
    for low, high, grp in groups:
        # find income changes by group
        sub_data = ati_data[
            (ati_data["base"] <= high) & (ati_data["base"] > low)
        ]
        results = find_percs(sub_data, grp)
        plot_data[grp] = results

    legend_labels = [
        "Increase of > 5%",
        "Increase 1-5%",
        "Change < 1%",
        "Decrease of 1-5%",
        "Decrease > 5%",
    ]
    labels = list(plot_data.keys())
    data = np.array(list(plot_data.values()))
    data_cumsum = data.cumsum(axis=1)
    category_colors = plt.get_cmap("GnBu")(
        np.linspace(0.15, 0.85, data.shape[1])
    )

    fig, ax = plt.subplots(figsize=figsize)
    ax.invert_yaxis()
    ax.set_xlim(0, np.sum(data, axis=1).max())

    for i, (colname, color) in enumerate(zip(legend_labels, category_colors)):
        widths = data[:, i]
        starts = data_cumsum[:, i] - widths
        ax.barh(
            labels, widths, left=starts, height=0.9, label=colname, color=color
        )
        if include_text:
            # add text label
            xcenters = starts + widths / 2
            r, g, b, _ = color
            text_color = "white" if r * g * b < 0.5 else "darkgrey"
            for y, (x, c) in enumerate(zip(xcenters, widths)):
                ax.text(
                    x,
                    y,
                    f"{c * 100:.1f}%",
                    ha="center",
                    va="center",
                    color=text_color,
                )
    ax.legend(bbox_to_anchor=(1, 1), loc="upper left", fontsize="small")
    ax.set_xlabel("Portion of Bin", fontweight="bold")
    ax.set_ylabel("Expanded Income Bin", fontweight="bold")
    ax.get_xaxis().set_major_formatter(
        mpl.ticker.FuncFormatter(lambda x, p: format(f"{int(x * 100)}%"))
    )
    if title == "default":
        title = f"Percentage Change In After Tax Income - {year}"
    ax.set_title(title)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.tick_params(axis="y", which="both", length=0, pad=15)

    return fig


def differences_plot(
    tb,
    tax_type: str,
    figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
    title: str = "default",
):
    """
    Create a bar chart that shows the change in total liability for a given
    tax

    Parameters
    ----------
    tb: TaxBrain object
        TaxBrain object for analysis
    tax_type: str
        tax for which to show the change in liability
        options: 'income', 'payroll', 'combined'
    figsize: tuple
        representing the size of the figure (width, height) in inches
    title: str
        title for plot

    Returns
    -------
    fig: Matplotlib.pyplot figure object
        differences plot
    """

    def axis_formatter(x, p):
        if x >= 0:
            return f"${x * 1e-9:,.2f}b"
        else:
            return f"-${x * 1e-9:,.2f}b"

    acceptable_taxes = ["income", "payroll", "combined"]
    msg = f"tax_type must be one of the following: {acceptable_taxes}"
    assert tax_type in acceptable_taxes, msg

    # find change in each tax variable
    tax_vars = ["iitax", "payrolltax", "combined"]
    agg_base = tb.multi_var_table(tax_vars, "base")
    agg_reform = tb.multi_var_table(tax_vars, "reform")
    agg_diff = agg_reform - agg_base

    # transpose agg_diff to make plotting easier
    plot_data = agg_diff.transpose()
    tax_var = tax_vars[acceptable_taxes.index(tax_type)]
    plot_data["color"] = np.where(plot_data[tax_var] < 0, "red", "blue")
    fig, ax = plt.subplots(figsize=figsize)
    ax.grid(True, axis="y", alpha=0.55)
    ax.set_axisbelow(True)
    ax.bar(
        plot_data.index,
        plot_data["combined"],
        alpha=0.55,
        color=plot_data["color"],
    )
    if title == "default":
        title = f"Change in Aggregate {tax_type.title()} Tax Liability"
    ax.set_title(title)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.get_yaxis().set_major_formatter(
        mpl.ticker.FuncFormatter(axis_formatter)
    )
    ax.xaxis.set_ticks(list(plot_data.index))
    ax.xaxis.set_major_formatter(mpl.ticker.ScalarFormatter(useOffset=False))

    return fig


def update_policy(
    policy_obj: tc.Policy,
    reform: Union[TaxcalcReform, ParamToolsAdjustment],
    **kwargs,
):
    """
    Convenience method that updates the Policy object with the reform
    dict using the appropriate method, given the reform format.

    Parameters
    ----------
    policy_obj: Tax-Calculator Policy class object
        Policy object for tax parameterization used for analysis
    reform: str or dict
        parameters for tax policy

    Returns
    -------
    None
        modifies the Policy object
    """
    if is_paramtools_format(reform):
        policy_obj.adjust(reform, **kwargs)
    else:
        policy_obj.implement_reform(reform, **kwargs)


def is_paramtools_format(reform: Union[TaxcalcReform, ParamToolsAdjustment]):
    """
    Check first item in reform to determine if it is using the ParamTools
    adjustment or the Tax-Calculator reform format.

    If first item is a dict, then it is likely be a Tax-Calculator reform:
        {
            param: {2020: 1000}
        }

    Otherwise, it is likely to be a ParamTools format.

    Parameters
    ----------
    reform: str or dict
        parameters for tax policy

    Returns
    -------
    bool: bool
        True if reform is likely to be in ParamTools format

    """
    for param, data in reform.items():
        if isinstance(data, dict):
            return False  # taxcalc reform
        else:
            # Not doing a specific check to see if the value is a list
            # since it could be a list or just a scalar value.
            return True


def lorenz_data(tb, year: int, var: str = "aftertax_income"):
    """
    Pull data used for the lorenz curve plot

    Parameters
    ----------
    tb: TaxBrain class object
        TaxBrain object for analysis
    year: int
        year of data to use
    var: str
        name of the variable to use

    Returns
    -------
    final_data: Pandas DataFrame
        DataFrame with Lorenz curve for baseline and reform
    """
    data = pd.DataFrame(
        {
            "base": tb.base_data[year][var],
            "reform": tb.reform_data[year][var],
            "wt": tb.base_data[year]["s006"],
        }
    )
    data["wt_base"] = data["base"] * data["wt"]
    data["wt_reform"] = data["reform"] * data["wt"]
    data.sort_values("base", inplace=True)
    data["cwt"] = data["wt"].cumsum()
    data["percentile"] = data["cwt"] / data["wt"].sum()
    # each bin has 1% of the population
    _bins = np.arange(0, 1.01, step=0.01)
    data["bin"] = pd.cut(data["percentile"], bins=_bins)
    gdf = data.groupby("bin", observed=False)
    base = gdf["wt_base"].sum()
    base = np.where(base < 0, 0, base)
    reform = gdf["wt_reform"].sum()
    reform = np.where(reform < 0, 0, reform)
    final_data = pd.DataFrame(
        {
            "Base": base.cumsum() / data["wt_base"].sum(),
            "Reform": reform.cumsum() / data["wt_reform"].sum(),
            "Population": gdf["wt"].sum().cumsum() / data["wt"].sum(),
        }
    )

    return final_data


def lorenz_curve(
    tb,
    year: int,
    var: str = "aftertax_income",
    figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
    xlabel: str = "Cummulative Percentage of Tax Units",
    ylabel: str = "Cummulative Percentage of Income",
    base_color: PlotColors = "blue",
    base_linestyle: str = "-",
    reform_color: PlotColors = "red",
    reform_linestyle: str = "--",
    dpi: Union[int, float] = 100,
):
    """
    Generate a Lorenz Curve

    Parameters
    ----------
    tb: TaxBrain class object
        TaxBrain object for analysis
    year: int
        year of data you want to use for the lorenz curve
    var: str
        name of the variable to use
    figsize: tuple
        representing the size of the figure (width, height) in inches
    xlabel: str
        x axis label
    ylabel: str
        y axis label
    base_color: str
        color used for the base line
    base_linestyle: str
        linestyle for the base line
    reform_color: str
        color used for the reform line
    reform_linestyle: str
        linestyle for the reform line
    dpi: int
        dots per inch in the figure image

    Returns
    -------
    None
    """
    plot_data = lorenz_data(tb, year, var)
    fig, ax = plt.subplots(figsize=figsize)
    ax.plot([0, 1], [0, 1], c="black", alpha=0.5)  # 45 degree line
    ax.plot(
        plot_data["Population"],
        plot_data["Base"],
        c=base_color,
        linestyle=base_linestyle,
        label="Base",
    )
    ax.plot(
        plot_data["Population"],
        plot_data["Reform"],
        c=reform_color,
        linestyle=reform_linestyle,
        label="Reform",
    )
    ax.legend(loc="upper left")
    ax.set_xlabel(xlabel, fontweight="bold")
    ax.set_ylabel(ylabel, fontweight="bold")
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    return fig


def volcano_plot(
    tb,
    year: int,
    y_var: str = "expanded_income",
    x_var: str = "combined",
    min_y: Union[int, float] = 0.01,
    max_y: Union[int, float] = 9e99,
    log_scale: bool = True,
    increase_color: PlotColors = "#F15FE4",
    decrease_color: PlotColors = "#41D6C2",
    dotsize: Union[int, float] = 0.75,
    alpha: float = 0.5,
    figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
    dpi: Union[int, float] = 100,
    xlabel: str = "Change in Tax Liability",
    ylabel: str = "Expanded Income",
):
    """
    Create a volcano plot to show change in tax tax liability

    Parameters
    ----------
    tb: TaxBrain class object
        TaxBrain object for analysis
    year: int
        year for the plot
    min_y: float
        minimum amount for the y variable to be included in the plot
    max_y: float
        maximum amount for the y variable to be included in the plot
    y_var: str
        variable on the y axis
    x_var: str
        variable on the x axis
    log_scale: bool
        whether the y-axis should use a log scale. If this is true,
        min_inc must be >= 0
    increase_color: str
        color to use for dots when x increases
    decrease_color: str
        color to use for dots when x decrease
    dotsize: int
        size of the dots in the scatter plot
    alpha: float
        attribute for transparency of the dots
    figsize: tuple
        the figure size of the plot (width, height) in inches
    dpi: int
        dots per inch in the figure
    xlabel: str
        label on the x axis
    ylabel: str
        label on the y axis

    Returns
    -------
    fig: Matplotlib.pyplot figure object
        volcano plot figure
    """

    def log_axis(x, pos):
        """
        Converts y-axis log values
        """
        return f"${np.exp(x):,.0f}"

    def axis_formatter(x, pos):
        if x >= 0:
            return f"${x:,.0f}"
        else:
            return f"-${abs(x):,.0f}"

    if log_scale and min_y < 0:
        msg = "`min_y` must be >= 0 when `log_scale` is true"
        raise ValueError(msg)
    _y = tb.base_data[year][y_var]
    _x_change = tb.reform_data[year][x_var] - tb.base_data[year][x_var]
    mask = np.logical_and(_y >= min_y, _y <= max_y)
    y = _y[mask]
    x_change = _x_change[mask]
    colors = [increase_color if x >= 0 else decrease_color for x in x_change]
    xformatter = ticker.FuncFormatter(axis_formatter)
    yformatter = ticker.FuncFormatter(axis_formatter)
    if log_scale:
        yformatter = ticker.FuncFormatter(log_axis)
        y = np.log(y)
    fig, ax = plt.subplots(figsize=figsize)
    ax.scatter(x_change, y, c=colors, s=dotsize, alpha=alpha)
    ax.axvline(0, color="black", alpha=0.5)
    ax.grid(True, linestyle="--")
    ax.xaxis.set_major_formatter(xformatter)
    ax.xaxis.set_tick_params(rotation=25)
    ax.yaxis.set_major_formatter(yformatter)
    ax.set_xlabel(xlabel, fontweight="bold")
    ax.set_ylabel(ylabel, fontweight="bold")

    return fig


def revenue_plot(
    tb,
    tax_vars: list = ["iitax", "payrolltax", "combined"],
    figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
):
    """Plot the changes in tax revenue from a given reform

    Parameters
    ----------
    tb : TaxBrain class object
        TaxBrain object for analysis
    tax_vars: list
        List of tax varaibles to include on the graph
    """

    def axis_formatter(x, p):
        if x >= 0:
            return f"${x * 1e-9:,.2f}"
        else:
            return f"-${x * 1e-9:,.2f}"

    assert tax_vars, "`tax_vars` must contain at least one tax variable"
    for var in tax_vars:
        if var not in ["iitax", "payrolltax", "combined"]:
            msg = (
                f"`{var}` is invalid. Valid tax variables are "
                "`iitax`, `payrolltax`, `combined`"
            )
            raise ValueError(msg)
    label_map = {
        "iitax": "Income",
        "payrolltax": "Payroll",
        "combined": "Combined",
    }
    color_map = {
        "Income: Base": "#12719e",
        "Income: Reform": "#73bfe2",
        "Payroll: Base": "#408941",
        "Payroll: Reform": "#98cf90",
        "Combined: Base": "#a4201d",
        "Combined: Reform": "#e9807d",
    }
    base_data = tb.multi_var_table(tax_vars, "base", include_total=False)
    reform_data = tb.multi_var_table(tax_vars, "reform", include_total=False)
    fig, ax = plt.subplots(figsize=figsize)
    years = base_data.columns
    for tax in tax_vars:
        base_label = f"{label_map[tax]}: Base"
        reform_label = f"{label_map[tax]}: Reform"
        ax.plot(
            years,
            base_data.loc[tax],
            label=base_label,
            color=color_map[base_label],
        )
        ax.plot(
            years,
            reform_data.loc[tax],
            label=reform_label,
            color=color_map[reform_label],
        )

    ax.legend(loc="upper right", bbox_to_anchor=(1.40, 1), title="Tax Type")
    ax.set_ylabel("Tax Liability (Billions)")
    ax.set_title("Tax Liability by Year")
    # remove plot borders
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    # convert y axis to billions
    ax.get_yaxis().set_major_formatter(
        mpl.ticker.FuncFormatter(axis_formatter)
    )
    return fig



from taxbrain import TaxBrain, report


reform_url = "https://raw.githubusercontent.com/PSLmodels/Tax-Calculator/master/taxcalc/reforms/Larson2019.json"

# run static analysis

tb_static = TaxBrain(2019, 2028, use_cps=True, reform=reform_url)
tb_static.run()
static_table = tb_static.weighted_totals("c00100")
print("Tax Liability by Year\n")
print("Static Results")
print(static_table)

# run dynamic analysis

tb_dynamic = TaxBrain(
    2019, 2028, use_cps=True, reform=reform_url, behavior={"sub": 0.25}
)
tb_dynamic.run()
dynamic_table = tb_dynamic.weighted_totals("c00100")
print("Dynamic Results")
print(dynamic_table)

# produce a differences table

diff = tb_static.differences_table(2019, "weighted_deciles", "combined")
print("\nDifferences Table for 2019")
print(diff)

# produce a distribution table

dist = tb_dynamic.distribution_table(
    2019, "weighted_deciles", "expanded_income", "reform"
)
print("\nDistribution Table for 2019")
print(dist)

# produce a pdf report summarizing the effects of the reform
outdir = "larsonreform"
name = "The Social Security 2100 Act: Rep. John Larson"
author = "Anderson Frailey"
report(tb_dynamic, name=name, author=author, outdir=outdir)



import os
import pytest
import pandas as pd
import numpy as np
from taxbrain import TaxBrain


def test_arg_validation():
    with pytest.raises(AssertionError):
        TaxBrain("2018", "2020", microdata="CPS")
    with pytest.raises(AssertionError):
        TaxBrain(
            TaxBrain.LAST_BUDGET_YEAR,
            TaxBrain.FIRST_BUDGET_YEAR,
            microdata="CPS",
        )
    with pytest.raises(AssertionError):
        TaxBrain(TaxBrain.FIRST_BUDGET_YEAR - 1, 2018, microdata="CPS")
    with pytest.raises(AssertionError):
        TaxBrain(2018, TaxBrain.LAST_BUDGET_YEAR + 1, microdata="CPS")


def test_static_run(tb_static):
    with pytest.raises(TypeError):
        tb_static.run(dict())
    tb_static.run()


def test_baseline_policy():
    base = {"II_em": {2019: 0}}
    reform = {"II_em": {2025: 2000}}

    tb = TaxBrain(2018, 2019, microdata="CPS", reform=reform, base_policy=base)
    tb.run()


@pytest.mark.requires_pufcsv
def test_baseline_policy_PUF():
    base = {"II_em": {2019: 0}}
    reform = {"II_em": {2025: 2000}}

    tb = TaxBrain(2018, 2019, microdata="PUF", reform=reform, base_policy=base)
    tb.run()


@pytest.mark.requires_tmdcsv
def test_baseline_policy_TMD():
    base = {"II_em": {2021: 0}}
    reform = {"II_em": {2025: 2000}}

    tb = TaxBrain(2021, 2022, microdata="TMD", reform=reform, base_policy=base)
    tb.run()


def test_run_corporate_distribution():
    """
    Tests the taxbrain.run method with the distribution of corporate
    income tax revenue.
    """
    base = {"II_em": {2019: 0}}
    reform = {"II_em": {2025: 2000}}
    corp_revenue = [1_000_000_000, 2_000_000_000]
    tb = TaxBrain(
        2018,
        2019,
        microdata="CPS",
        reform=reform,
        base_policy=base,
        corp_revenue=corp_revenue,
    )
    tb.run()


def test_dynamic_run_corporate_distribution():
    """
    Tests the taxbrain.run method with the distribution of corporate
    income tax revenue.
    """
    base = {"II_em": {2019: 0}}
    reform = {"II_em": {2025: 2000}}
    corp_revenue = [1_000_000_000, 2_000_000_000]
    tb = TaxBrain(
        2018,
        2019,
        microdata="CPS",
        reform=reform,
        base_policy=base,
        behavior={"sub": 0.25},
        corp_revenue=corp_revenue,
    )
    tb.run()


def test_dynamic_run(tb_dynamic):
    tb_dynamic.run()


def test_stacked_run():
    # reforms to use
    payroll_json = """{"SS_Earnings_thd": {"2021": 400000}}"""
    CG_rate_json = """{
        "CG_brk3": {"2021": [1000000, 1000000, 1000000, 1000000, 1000000]},
        "CG_rt4": {"2021": 0.396}
    }"""
    reform_dict = {
        "Payroll Threshold Increase": payroll_json,
        "Capital Gains Tax Changes": CG_rate_json,
    }
    tb = TaxBrain(
        2021, 2022, reform=reform_dict, stacked=True, microdata="CPS"
    )
    tb.run()
    # check that there is a stacked table now
    assert isinstance(tb.stacked_table, pd.DataFrame)


def test_stacked_run_corporate():
    # reforms to use
    payroll_json = """{"SS_Earnings_thd": {"2021": 400000}}"""
    CG_rate_json = """{
        "CG_brk3": {"2021": [1000000, 1000000, 1000000, 1000000, 1000000]},
        "CG_rt4": {"2021": 0.396}
    }"""
    reform_dict = {
        "Payroll Threshold Increase": payroll_json,
        "Capital Gains Tax Changes": CG_rate_json,
    }
    tb = TaxBrain(
        2021,
        2022,
        reform=reform_dict,
        stacked=True,
        microdata="CPS",
        corp_revenue=[100_000_000, 100_000_000],
    )
    tb.run()
    # check that there is a stacked table now
    assert isinstance(tb.stacked_table, pd.DataFrame)


def test_weighted_totals(tb_static):
    tb_static.run()
    table = tb_static.weighted_totals("combined")
    assert isinstance(table, pd.DataFrame)
    # table.to_csv("expected_weighted_table.csv")
    cur_path = os.path.dirname(os.path.abspath(__file__))
    expected_table = pd.read_csv(
        os.path.join(cur_path, "expected_weighted_table.csv"), index_col=0
    )
    # convert columns to integers to avoid a meaningless error
    expected_table.columns = [int(col) for col in expected_table.columns]
    diffs = False
    for col in table.columns:
        if not np.allclose(table[col], expected_table[col]):
            diffs = True
    if diffs:
        new_file_name = "actual_weighted_table.csv"
        table.to_csv(os.path.join(cur_path, new_file_name))
        msg = (
            "Weighted table results differ from expected. New results are"
            " in actual_weighted_table.csv. If new results are ok, copy"
            " actual_weighted_table.csv to expected_weighted_table.csv"
            " and rerun test."
        )
        raise ValueError(msg)


def test_multi_var_table(tb_dynamic):
    tb_dynamic.run()
    with pytest.raises(ValueError):
        tb_dynamic.multi_var_table(["iitax"], "calc")
    with pytest.raises(TypeError):
        tb_dynamic.multi_var_table(set(), "base")
    table = tb_dynamic.multi_var_table(
        ["iitax", "payrolltax", "combined"], "reform"
    )
    assert isinstance(table, pd.DataFrame)


def test_differences_table(tb_dynamic):
    tb_dynamic.run()
    table = tb_dynamic.differences_table(2018, "weighted_deciles", "combined")
    assert isinstance(table, pd.DataFrame)


def test_distribution_table(tb_static):
    tb_static.run()
    table = tb_static.distribution_table(
        2019, "weighted_deciles", "expanded_income_baseline", "reform"
    )
    assert isinstance(table, pd.DataFrame)
    table = tb_static.distribution_table(
        2019, "weighted_deciles", "expanded_income", "base"
    )
    assert isinstance(table, pd.DataFrame)
    with pytest.raises(ValueError):
        tb_static.distribution_table(
            2018, "weighted_deciles", "expanded_income", "nonreform"
        )


def test_user_input(reform_json_str, assump_json_str):
    valid_reform = {"II_rt7": {2019: 0.40}}
    # Test valid reform dictionary with No assumption
    TaxBrain(2018, 2020, microdata="CPS", reform=valid_reform)
    TaxBrain(2018, 2020, microdata="CPS", reform=reform_json_str)
    invalid_assump = {"consumption": {}}
    # Test valid reform and assumptions dictionary
    valid_assump = {
        "consumption": {},
        "growdiff_baseline": {},
        "growdiff_response": {},
    }
    TaxBrain(2018, 2019, microdata="CPS", assump=valid_assump)
    TaxBrain(
        2018,
        2019,
        microdata="CPS",
        reform=reform_json_str,
        assump=assump_json_str,
    )
    tb = TaxBrain(
        2018, 2019, microdata="CPS", reform=valid_reform, assump=valid_assump
    )
    required_param_keys = {
        "policy",
        "consumption",
        "growdiff_baseline",
        "growdiff_response",
        "behavior",
        "base_policy",
    }
    assert set(tb.params.keys()) == required_param_keys
    with pytest.raises(ValueError):
        TaxBrain(2018, 2020, microdata="CPS", assump=invalid_assump)
    invalid_assump = {
        "consumption": {},
        "growdiff_baseline": {},
        "growdiff_response": {},
        "invalid": {},
    }
    with pytest.raises(ValueError):
        TaxBrain(2018, 2020, microdata="CPS", assump=invalid_assump)
    with pytest.raises(TypeError):
        TaxBrain(2018, 2020, microdata="CPS", reform=True)
    with pytest.raises(TypeError):
        TaxBrain(2018, 2020, microdata="CPS", assump=True)



import taxcalc as tc
import pandas as pd
import numpy as np
import behresp
from taxcalc.utils import (
    DIST_VARIABLES,
    DIFF_VARIABLES,
    create_distribution_table,
    create_difference_table,
)
from dask import compute, delayed
import dask.multiprocessing
from collections import defaultdict
from taxbrain.utils import weighted_sum, update_policy
from taxbrain.corporate_incidence import distribute as dist_corp
from typing import Union
from paramtools import ValidationError
from pathlib import Path


class TaxBrain:

    FIRST_BUDGET_YEAR = tc.Policy.JSON_START_YEAR
    LAST_BUDGET_YEAR = tc.Policy.LAST_BUDGET_YEAR
    # Default list of variables saved for each year
    DEFAULT_VARIABLES = list(set(DIST_VARIABLES).union(set(DIFF_VARIABLES)))

    # add dictionary to hold version of the various models
    VERSIONS = {
        "Tax-Calculator": tc.__version__,
        "Behavioral-Responses": behresp.__version__,
    }

    # add expected TMD filenames as constants
    TMD_DATA_FILE = "tmd.csv.gz"
    TMD_WEIGHTS_FILE = "tmd_weights.csv.gz"
    TMD_GROWFACTORS_FILE = "tmd_growfactors.csv"

    def __init__(
        self,
        start_year: int,
        end_year: int = LAST_BUDGET_YEAR,
        microdata: Union[str, dict] = "CPS",
        reform: Union[str, dict] = None,
        behavior: dict = None,
        assump=None,
        base_policy: Union[str, dict] = None,
        corp_revenue: Union[dict, list, np.array] = None,
        corp_incidence_assumptions: dict = None,
        verbose=False,
        stacked=False,
    ):
        """
        Constructor for the TaxBrain class

        Parameters
        ----------
        start_year: int
            First year in the analysis. Must be no earlier than the
            first year allowed in Tax-Calculator.
        end_year: int
            Last year in the analysis. Must be no later than the last
            year allowed in Tax-Calculator.
        microdata: str or dict
            A string in ["CPS", "PUF", "TMD"] or path to a micro-data
            file or a Pandas DataFrame, containing micro-data, or a
            dictionary containing a path to microdata, associated
            weights, and grow factors.  If a dict, must have keys:
            "data", "start_year", "growfactors", "weights"
        reform: str or dict
            Individual income tax policy reform. Can be either a string
            pointing to a JSON reform file, or the contents of a JSON file,
            or a properly formatted JSON file.
        behavior: dict
            Individual behavior assumptions use by the Behavior-Response
            package.
        assump: str
            A string pointing to a JSON file containing user specified
            economic assumptions.
        base_policy: str or dict
            Individual income tax policy to use as the baseline for
            the analysis. This policy will be implemented in the base
            calculator instance as well as the reform Calculator
            before the user provided reform is implemented. Can either
            be a string pointing to a JSON reform file, the contents
            of a JSON file, or a properly formatted dictionary.
        corp_revenue: dict, list, or numpy array
            A set of corporate revenue estimates for a given set of
            years.  The estimates much line up with start_year and
            end_year.
        corp_incidence_assumptions: dict
            A dictionary summarizing the assumptions about the
            distribution of the corporate income tax.  See
            taxbrain.corporate_incidence.CI_params for an example.
        verbose: bool
            A boolean value indicated whether or not to write model
            progress reports.
        stacked: bool
            A boolean value indicating weather the provided reform is in the
            format used for stacked reform analysis

        Returns
        -------
        None
        """
        assert isinstance(start_year, int) & isinstance(
            end_year, int
        ), "Start and end years must be integers"
        assert start_year <= end_year, (
            f"Specified end year, {end_year}, is before specified start year, "
            f"{start_year}."
        )
        assert TaxBrain.FIRST_BUDGET_YEAR <= start_year, (
            f"Specified start_year, {start_year}, comes before first known "
            f"budget year, {TaxBrain.FIRST_BUDGET_YEAR}."
        )
        assert end_year <= TaxBrain.LAST_BUDGET_YEAR, (
            f"Specified end_year, {end_year}, comes after last known "
            f"budget year, {TaxBrain.LAST_BUDGET_YEAR}."
        )
        if corp_revenue:
            assert (
                len(corp_revenue) == end_year - start_year + 1
            ), f"Corporate revenue is not given for each budget year"
        self.microdata = microdata
        self.start_year = start_year
        self.end_year = end_year
        self.base_data = {yr: {} for yr in range(start_year, end_year + 1)}
        self.reform_data = {yr: {} for yr in range(start_year, end_year + 1)}
        self.corp_revenue = corp_revenue
        self.ci_params = corp_incidence_assumptions
        self.verbose = verbose
        self.stacked = stacked
        self.stacked_reforms = None  # only used if stacked is true

        # Process user inputs early to throw any errors quickly
        self.params = self._process_user_mods(reform, assump)
        self.params["behavior"] = behavior
        if base_policy:
            base_policy = self._process_user_mods(base_policy, None)
            self.params["base_policy"] = base_policy["policy"]
        else:
            self.params["base_policy"] = None

        self.has_run = False

    def run(
        self, varlist: list = DEFAULT_VARIABLES, client=None, num_workers=1
    ):
        """
        Run the calculators. TaxBrain will determine whether to do a static or
        partial equilibrium run based on the user's inputs when initializing
        the TaxBrain object.

        Parameters
        ----------
        varlist: list
            variables from the microdata to be stored in each year

        Returns
        -------
        None
        """
        if not isinstance(varlist, list):
            msg = f"'varlist' is of type {type(varlist)}. Must be a list."
            raise TypeError(msg)
        if self.stacked:
            base_calc, policy, records = self._make_stacked_objects()
            self._stacked_run(
                varlist, base_calc, policy, records, client, num_workers
            )
            del base_calc
        else:
            base_calc, reform_calc = self._make_calculators()
            if self.params["behavior"]:
                if self.verbose:
                    print("Running dynamic simulations")
                self._dynamic_run(
                    varlist, base_calc, reform_calc, client, num_workers
                )
            else:
                if self.verbose:
                    print("Running static simulations")
                self._static_run(
                    varlist, base_calc, reform_calc, client, num_workers
                )
            del base_calc, reform_calc

        setattr(self, "has_run", True)

    def weighted_totals(
        self, var: str, include_total: bool = False
    ) -> pd.DataFrame:
        """
        Create a pandas DataFrame that shows the weighted sum or a specified
        variable under the baseline policy, reform policy, and the difference
        between the two.

        Parameters
        ----------
        var: str
            Variable name for variable you want the weighted total of.
        include_total: bool
            If true the returned DataFrame will include a "total" columns

        Returns
        -------
        Pandas DataFrame
            A Pandas DataFrame with rows for the baseline total,
            reform total, and the difference between the two.
        """
        base_totals = {}
        reform_totals = {}
        differences = {}
        for year in range(self.start_year, self.end_year + 1):
            base_totals[year] = (
                self.base_data[year]["s006"] * self.base_data[year][var]
            ).sum()
            reform_totals[year] = (
                self.reform_data[year]["s006"] * self.reform_data[year][var]
            ).sum()
            differences[year] = reform_totals[year] - base_totals[year]
        table = pd.DataFrame(
            [base_totals, reform_totals, differences],
            index=["Base", "Reform", "Difference"],
        )
        if include_total:
            table["Total"] = table.sum(axis=1)
        return table

    def multi_var_table(
        self, varlist: list, calc: str, include_total: bool = False
    ) -> pd.DataFrame:
        """
        Create a Pandas DataFrame with multiple variables from the
        specified data source

        Parameters
        ----------
        varlist: list
            list of variables to include in the table
        calc: str
            specify reform or base calculator data, can take either
            `'REFORM'` or `'BASE'`
        include_total: bool
            If true the returned DataFrame will include a "total" column

        Returns
        -------
        df: Pandas DataFrame
            A Pandas DataFrame containing the weighted sum of each
            variable passed in the `varlist` argument for each year in
            the analysis.
        """
        if not isinstance(varlist, list):
            msg = f"'varlist' is of type {type(varlist)}. Must be a list."
            raise TypeError(msg)
        if calc.upper() == "REFORM":
            data = self.reform_data
        elif calc.upper() == "BASE":
            data = self.base_data
        else:
            raise ValueError("'calc' must be 'base' or 'reform'")
        data_dict = defaultdict(list)
        for year in range(self.start_year, self.end_year + 1):
            for var in varlist:
                data_dict[var] += [weighted_sum(data[year], var)]
        df = pd.DataFrame(
            data_dict, index=range(self.start_year, self.end_year + 1)
        )
        table = df.transpose()
        if include_total:
            table["Total"] = table.sum(axis=1)
        return table

    def distribution_table(
        self,
        year: int,
        groupby: str,
        income_measure: str,
        calc: str,
        pop_quantiles: bool = False,
    ) -> pd.DataFrame:
        """
        Method to create a distribution table

        Parameters
        ----------
        year: int
            which year the distribution table data should be from
        groupby: str
            determines how the rows in the table are sorted
            options: 'weighted_deciles', 'standard_income_bins',
            'soi_agi_bin'
        income_measure: str
            determines which variable is used to sort the rows in
            the table
            options: 'expanded_income' or 'expanded_income_baseline'
        calc: str
            which calculator to use, can take either
            `'REFORM'` or `'BASE'`
        calc: which calculator to use: base or reform
        pop_quantiles: bool
            whether or not weighted_deciles contain equal number of
            tax units (False) or people (True)

        Returns
        -------
        table: Pandas DataFrame
            distribution table
        """
        # pull desired data
        if calc.lower() == "base":
            data = self.base_data[year]
        elif calc.lower() == "reform":
            data = self.reform_data[year]
        else:
            raise ValueError("calc must be either BASE or REFORM")
        # minor data preparation before calling the function
        if pop_quantiles:
            data["count"] = data["s006"] * data["XTOT"]
        else:
            data["count"] = data["s006"]
        data["count_ItemDed"] = data["count"].where(data["c04470"] > 0.0, 0.0)
        data["count_StandardDed"] = data["count"].where(
            data["standard"] > 0.0, 0.0
        )
        data["count_AMT"] = data["count"].where(data["c09600"] > 0.0, 0.0)
        if income_measure == "expanded_income_baseline":
            base_income = self.base_data[year]["expanded_income"]
            data["expanded_income_baseline"] = base_income
        table = create_distribution_table(
            data, groupby, income_measure, pop_quantiles
        )
        return table

    def differences_table(
        self,
        year: int,
        groupby: str,
        tax_to_diff: str,
        pop_quantiles: bool = False,
    ) -> pd.DataFrame:
        """
        Method to create a differences table

        Parameters
        ----------
        year: int
            which year the difference table should be from
        groupby: str
            determines how the rows in the table are sorted
            options: 'weighted_deciles', 'standard_income_bins', 'soi_agi_bin'
        tax_to_diff: str
            which tax to take the difference of
            options: 'iitax', 'payrolltax', 'combined'
        pop_quantiles: bool
            whether weighted_deciles contain an equal number of tax
            units (False) or people (True)

        Returns
        -------
        table: Pandas DataFrame
            differences table
        """
        base_data = self.base_data[year]
        reform_data = self.reform_data[year]
        table = create_difference_table(
            base_data, reform_data, groupby, tax_to_diff, pop_quantiles
        )
        return table

    # ----- private methods -----
    def _taxcalc_advance(self, calc, varlist, year, reform=False):
        """
        This function advances the year used in Tax-Calculator, computes
        tax liability and rates, and saves the results to a dictionary.
        Args:
            calc (Tax-Calculator Calculator object): TC calculator
            varlist (list): variables to return
            year (int): year to begin advancing from
            reform (bool): whether Calculator object is for the reform policy

        Returns:
            tax_dict (dict): a dictionary of microdata with marginal tax
                rates and other information computed in TC
        """
        calc.advance_to_year(year)
        if self.corp_revenue is not None:
            if reform:
                calc = dist_corp(
                    calc,
                    self.corp_revenue,
                    year,
                    self.start_year,
                    self.ci_params,
                )
        calc.calc_all()
        df = calc.dataframe(varlist)

        return df

    def _behresp_advance(self, base_calc, reform_calc, varlist, year):
        """
        This function advances the year used in the Behavioral Responses
        model and saves the results to a dictionary.
        Args:
            calc1 (Tax-Calculator Calculator object): TC calculator
            year (int): year to begin advancing from
        Returns:
            tax_dict (dict): a dictionary of microdata with marginal tax
                rates and other information computed in TC
        """
        base_calc.advance_to_year(year)
        reform_calc.advance_to_year(year)
        if self.corp_revenue is not None:
            reform_calc = dist_corp(
                reform_calc,
                self.corp_revenue,
                year,
                self.start_year,
                self.ci_params,
            )
        base, reform = behresp.response(
            base_calc, reform_calc, self.params["behavior"], dump=True
        )
        base_df = base[varlist]
        reform_df = reform[varlist]

        return [base_df, reform_df]

    def _static_run(
        self, varlist, base_calc, reform_calc, client, num_workers
    ):
        """
        Run the calculator for a static analysis
        """
        if "s006" not in varlist:  # ensure weight is always included
            varlist.append("s006")
        lazy_values = []
        for yr in range(self.start_year, self.end_year + 1):
            lazy_values.extend(
                [
                    delayed(self._taxcalc_advance(base_calc, varlist, yr)),
                    delayed(
                        self._taxcalc_advance(
                            reform_calc, varlist, yr, reform=True
                        )
                    ),
                ]
            )
        if client:
            futures = client.compute(lazy_values, num_workers=num_workers)
            results = client.gather(futures)
        else:
            results = results = compute(
                *lazy_values,
                scheduler=dask.multiprocessing.get,
                num_workers=num_workers,
            )

        # add results to base and reform data
        yr = self.start_year
        for i in np.arange(0, len(results), 2):
            self.base_data[yr] = results[i]
            self.reform_data[yr] = results[i + 1]
            yr += 1

        del results

    def _dynamic_run(
        self, varlist, base_calc, reform_calc, client, num_workers
    ):
        """
        Run a dynamic response
        """
        if "s006" not in varlist:  # ensure weight is always included
            varlist.append("s006")
        lazy_values = []
        for yr in range(self.start_year, self.end_year + 1):
            lazy_values.extend(
                [
                    delayed(
                        self._behresp_advance(
                            base_calc, reform_calc, varlist, yr
                        )
                    )
                ]
            )
        if client:
            futures = client.compute(lazy_values, num_workers=num_workers)
            results = client.gather(futures)
        else:
            results = results = compute(
                *lazy_values,
                scheduler=dask.multiprocessing.get,
                num_workers=num_workers,
            )

        # add results to base and reform data
        for i in range(len(results)):
            yr = self.start_year + i
            self.base_data[yr] = results[i][0]
            self.reform_data[yr] = results[i][1]

        del results

    def _stacked_run(
        self, varlist, base_calc, policy, records, client, num_workers
    ):
        revenue_output = {}
        BW_len = self.end_year - self.start_year + 1
        # run the base calc first to get baseline results
        lazy_values = []
        for yr in range(self.start_year, self.end_year + 1):
            lazy_values.append(
                delayed(self._taxcalc_advance(base_calc, varlist, yr))
            )
        if client:
            futures = client.compute(lazy_values, num_workers=num_workers)
            results = client.gather(futures)
        else:
            results = results = compute(
                *lazy_values,
                scheduler=dask.multiprocessing.get,
                num_workers=num_workers,
            )
        # add results to data and revenue outputs
        revenue_output["Baseline"] = np.zeros(BW_len)
        yr = self.start_year
        # for i in np.arange(0, len(results), 2):
        for i, res in enumerate(results):
            self.base_data[yr] = res
            combined = (res["combined"] * res["s006"]).sum()
            revenue_output["Baseline"][yr - self.start_year] = combined
            yr += 1
        reform_list = list(self.stacked_reforms.keys())
        # Loop over different provisions
        for k, v in self.stacked_reforms.items():
            if self.verbose:
                print("Analyzing ", k)
            revenue_output[k] = np.zeros(BW_len)
            ref = policy.read_json_reform(v)
            # update Policy object with additional provisions
            policy.implement_reform(ref)
            # update Calculator object with new Policy object
            calc = tc.calculator.Calculator(policy=policy, records=records)
            # loop over each year in budget window
            for yr in np.arange(self.start_year, self.end_year + 1):
                calc.advance_to_year(yr)
                # change income in accordance with corp income tax
                # distributed across individual taxpayers
                if self.corp_revenue is not None:
                    calc = dist_corp(
                        calc,
                        self.corp_revenue,
                        yr,
                        self.start_year,
                        self.ci_params,
                    )
                # makes calculations on microdata
                calc.calc_all()
                # compute total revenue
                revenue_output[k][yr - self.start_year] = calc.weighted_total(
                    "combined"
                )
                # if we're on the last reform piece, save the data
                if k == reform_list[-1]:
                    self.reform_data[yr] = calc.dataframe(varlist)
        df = pd.DataFrame.from_dict(
            revenue_output,
            orient="Index",
            columns=np.arange(self.start_year, self.start_year + BW_len),
        )
        # Compute differences from one provision to another
        rev_est_tbl = df.diff()
        # Drop baseline revenue since reporting differences relative to baseline
        rev_est_tbl.drop(labels="Baseline", inplace=True)

        # Create totals across budget window
        tot_col = f"{self.start_year}-{self.end_year}"
        rev_est_tbl[tot_col] = rev_est_tbl[list(rev_est_tbl.columns)].sum(
            axis=1
        )
        # Create totals across provisions
        rev_est_tbl.loc["Total"] = rev_est_tbl.sum()

        # save the table as an attribute of the TaxBrain object
        setattr(self, "stacked_table", rev_est_tbl)

    def _process_user_mods(self, reform, assump):
        """
        Logic to process user mods and set self.params
        """

        def key_validation(actual_keys, required_keys, d_name):
            """
            Validate keys if reform or assump is passed as a dictionary
            """
            missing_keys = required_keys - actual_keys
            if missing_keys:
                msg = f"Required key(s) {missing_keys} missing from '{d_name}'"
                raise ValueError(msg)
            illegal_keys = actual_keys - required_keys
            if illegal_keys:
                msg = f"Illegal key(s) {illegal_keys} found in '{d_name}'"
                raise ValueError(msg)

        if isinstance(reform, dict):
            # If the reform is a dictionary, we'll leave it to Tax-Calculator
            # to catch errors in its implementation. Or, it's a stacked reform
            # and we will check that separately
            if isinstance(assump, str) or not assump:
                params = tc.Calculator.read_json_param_objects(None, assump)
            elif isinstance(assump, dict):
                actual_keys = set(assump.keys())
                required_keys = tc.Calculator.REQUIRED_ASSUMP_KEYS
                key_validation(actual_keys, required_keys, "assump")
                params = {**assump}
            else:
                raise TypeError(
                    "'assump' is not a string, dictionary, or None"
                )
            params["policy"] = reform
            # Check stacked reforms
            if self.stacked:
                pol = tc.Policy()
                full_policy = {}
                for name, subreform in reform.items():
                    # assume that if the reform is a string it's a JSON reform
                    # Otherwise the parameters are already a dictionary
                    if isinstance(subreform, str):
                        pol_params = tc.Policy.read_json_reform(subreform)
                    else:
                        m = "Reform must be valid JSON string or dictionary"
                        assert isinstance(subreform, dict), m
                        pol_params = subreform
                    try:
                        update_policy(pol, pol_params)
                    except ValidationError as e:
                        print(f"Validation error in {name}")
                        raise e
                    # update the full policy to save later
                    full_policy = {**full_policy, **pol_params}
                setattr(self, "stacked_reforms", reform)
                params["policy"] = full_policy
        elif isinstance(reform, str) or not reform:
            if isinstance(assump, str) or not assump:
                params = tc.Calculator.read_json_param_objects(reform, assump)
            elif isinstance(assump, dict):
                # Check to ensure that the assumption dictionary contains
                # all the needed keys. Tax-Calculator will check that they
                # are ultimately defined correctly when attempting to
                # use them.
                actual_keys = set(assump.keys())
                required_keys = tc.Calculator.REQUIRED_ASSUMP_KEYS
                key_validation(actual_keys, required_keys, "assump")
                params = tc.Calculator.read_json_param_objects(reform, None)
                for key in assump.keys():
                    params[key] = assump[key]
            else:
                raise TypeError(
                    "'assump' is not a string, dictionary, or None"
                )
        else:
            raise TypeError("'reform' is not a string, dictionary, or None")

        # confirm that all the expected keys are there
        required_keys = (
            tc.Calculator.REQUIRED_ASSUMP_KEYS
            | tc.Calculator.REQUIRED_REFORM_KEYS
        )
        assert set(params.keys()) == required_keys

        return params

    def _make_calculators(self):
        """
        This function creates the baseline and reform calculators used when
        the `run()` method is called
        """
        # Create two microsimulation calculators
        # Baseline calculator
        gd_base = tc.GrowDiff()
        gf_base = tc.GrowFactors()
        # apply user specified growdiff
        if self.params["growdiff_baseline"]:
            gd_base.update_growdiff(self.params["growdiff_baseline"])
            gd_base.apply_to(gf_base)
        if self.microdata == "CPS":
            records = tc.Records.cps_constructor(data=None, gfactors=gf_base)
        elif self.microdata == "PUF":
            records = tc.Records.puf_constructor(
                data="puf.csv",
                gfactors=gf_base,
                # weights=tc.Records.PUF_WEIGHTS_FILENAME,
            )
        elif self.microdata == "TMD":
            gf_base = tc.GrowFactors(self.TMD_GROWFACTORS_FILE)
            if self.params["growdiff_baseline"]:
                gd_base.apply_to(gf_base)

            records = tc.Records.tmd_constructor(
                data_path=Path(self.TMD_DATA_FILE),
                weights_path=Path(self.TMD_WEIGHTS_FILE),
                growfactors=gf_base,
            )
        elif isinstance(self.microdata, dict):
            if self.microdata["growfactors"] is None:
                gd_base = tc.GrowDiff()
                gf_base = tc.GrowFactors()
                # apply user specified growdiff
                if self.params["growdiff_baseline"]:
                    gd_base.update_growdiff(self.params["growdiff_baseline"])
                    gd_base.apply_to(gf_base)
            else:
                gd_base = tc.GrowDiff()
                gf_base = tc.GrowFactors(self.microdata["growfactors"])
                # apply user specified growdiff
                if self.params["growdiff_baseline"]:
                    gd_base.update_growdiff(self.params["growdiff_baseline"])
                    gd_base.apply_to(gf_base)
            records = tc.Records(
                self.microdata["data"],
                start_year=self.microdata["start_year"],
                gfactors=gf_base,
                weights=self.microdata["weights"],
            )
        else:
            raise ValueError(
                "microdata must be 'CPS', 'PUF', 'TMD', or a dictionary"
            )
        policy = tc.Policy(gf_base)
        if self.params["base_policy"]:
            update_policy(policy, self.params["base_policy"])
        base_calc = tc.Calculator(
            policy=policy, records=records, verbose=self.verbose
        )

        # Reform calculator
        gd_reform = tc.GrowDiff()
        gf_reform = tc.GrowFactors()
        # apply user specified growdiff
        if self.params["growdiff_response"]:
            gd_reform.update_growdiff(self.params["growdiff_response"])
            gd_reform.apply_to(gf_reform)
        if self.microdata == "CPS":
            records = tc.Records.cps_constructor(data=None, gfactors=gf_reform)
        elif self.microdata == "PUF":
            records = tc.Records.puf_constructor(
                data="puf.csv",
                gfactors=gf_reform,
                # weights=tc.Records.PUF_WEIGHTS_FILENAME,
            )
        elif self.microdata == "TMD":
            gf_reform = tc.GrowFactors(self.TMD_GROWFACTORS_FILE)
            if self.params["growdiff_response"]:
                gd_reform.apply_to(gf_reform)

            records = tc.Records.tmd_constructor(
                data_path=Path(self.TMD_DATA_FILE),
                weights_path=Path(self.TMD_WEIGHTS_FILE),
                growfactors=gf_reform,
            )
        elif isinstance(self.microdata, dict):
            if self.microdata["growfactors"] is None:
                gd_reform = tc.GrowDiff()
                gf_reform = tc.GrowFactors()
                # apply user specified growdiff
                if self.params["growdiff_response"]:
                    gd_reform.update_growdiff(self.params["growdiff_response"])
                    gd_reform.apply_to(gf_reform)
            else:
                gd_reform = tc.GrowDiff()
                gf_reform = tc.GrowFactors(self.microdata["growfactors"])
                # apply user specified growdiff
                if self.params["growdiff_response"]:
                    gd_reform.update_growdiff(self.params["growdiff_response"])
                    gd_reform.apply_to(gf_reform)
            records = tc.Records(
                self.microdata["data"],
                start_year=self.microdata["start_year"],
                gfactors=gf_reform,
                weights=self.microdata["weights"],
            )
        else:
            raise ValueError(
                "microdata must be 'CPS', 'PUF', 'TMD', or a dictionary"
            )
        policy = tc.Policy(gf_reform)
        if self.params["base_policy"]:
            update_policy(policy, self.params["base_policy"])
        update_policy(policy, self.params["policy"])

        # Initialize Calculator
        reform_calc = tc.Calculator(
            policy=policy, records=records, verbose=self.verbose
        )
        # delete all unneeded variables
        del gd_base, gd_reform, records, gf_base, gf_reform, policy
        return base_calc, reform_calc

    # TODO: update these method to allow for different microdata as above
    # but code becoming cumbersome, so should probably streamline in a single get calculator function
    def _make_stacked_objects(self):
        """
        This method makes the base calculator and policy and records objects
        for stacked reforms. The difference between this and the standard
        _make_calcuators method is that this method
        only fully creates the baseline calculator. For the reform, it creates
        policy and records objects and implements any growth assumptions
        provided by the user.
        """
        # Create two microsimulation calculators
        gd_base = tc.GrowDiff()
        gf_base = tc.GrowFactors()
        # apply user specified growdiff
        if self.params["growdiff_baseline"]:
            gd_base.update_growdiff(self.params["growdiff_baseline"])
            gd_base.apply_to(gf_base)
        if self.microdata == "CPS":
            records = tc.Records.cps_constructor(data=None, gfactors=gf_base)
        elif self.microdata == "PUF":
            records = tc.Records.puf_constructor(
                data="puf.csv",
                gfactors=gf_base,
                # weights=tc.Records.PUF_WEIGHTS_FILENAME,
            )
        elif self.microdata == "TMD":
            gf_base = tc.GrowFactors(self.TMD_GROWFACTORS_FILE)
            if self.params["growdiff_baseline"]:
                gd_base.apply_to(gf_base)

            records = tc.Records.tmd_constructor(
                data_path=Path(self.TMD_DATA_FILE),
                weights_path=Path(self.TMD_WEIGHTS_FILE),
                growfactors=gf_base,
            )
        elif isinstance(self.microdata, dict):
            if self.microdata["growfactors"] is None:
                gd_base = tc.GrowDiff()
                gf_base = tc.GrowFactors()
                # apply user specified growdiff
                if self.params["growdiff_baseline"]:
                    gd_base.update_growdiff(self.params["growdiff_baseline"])
                    gd_base.apply_to(gf_base)
            else:
                gd_base = tc.GrowDiff()
                gf_base = tc.GrowFactors(self.microdata["growfactors"])
                # apply user specified growdiff
                if self.params["growdiff_baseline"]:
                    gd_base.update_growdiff(self.params["growdiff_baseline"])
                    gd_base.apply_to(gf_base)
            records = tc.Records(
                self.microdata["data"],
                start_year=self.microdata["start_year"],
                gfactors=gf_base,
                weights=self.microdata["weights"],
            )
        else:
            raise ValueError(
                "microdata must be 'CPS', 'PUF', 'TMD', or a dictionary"
            )
        policy = tc.Policy(gf_base)
        if self.params["base_policy"]:
            update_policy(policy, self.params["base_policy"])
        base_calc = tc.Calculator(
            policy=policy, records=records, verbose=self.verbose
        )

        # Reform calculator
        gd_reform = tc.GrowDiff()
        gf_reform = tc.GrowFactors()
        # apply user specified growdiff
        if self.params["growdiff_response"]:
            gd_reform.update_growdiff(self.params["growdiff_response"])
            gd_reform.apply_to(gf_reform)
        if self.microdata == "CPS":
            reform_records = tc.Records.cps_constructor(
                data=None, gfactors=gf_reform
            )
        elif self.microdata == "PUF":
            reform_records = tc.Records.puf_constructor(
                data="puf.csv",
                gfactors=gf_reform,
                # weights=tc.Records.PUF_WEIGHTS_FILENAME,
            )
        elif self.microdata == "TMD":
            gf_reform = tc.GrowFactors(self.TMD_GROWFACTORS_FILE)
            if self.params["growdiff_response"]:
                gd_reform.apply_to(gf_reform)

            reform_records = tc.Records.tmd_constructor(
                data_path=Path(self.TMD_DATA_FILE),
                weights_path=Path(self.TMD_WEIGHTS_FILE),
                growfactors=gf_reform,
            )
        elif isinstance(self.microdata, dict):
            if self.microdata["growfactors"] is None:
                gd_reform = tc.GrowDiff()
                gf_reform = tc.GrowFactors()
                # apply user specified growdiff
                if self.params["growdiff_response"]:
                    gd_reform.update_growdiff(self.params["growdiff_response"])
                    gd_reform.apply_to(gf_reform)
            else:
                gd_reform = tc.GrowDiff()
                gf_reform = tc.GrowFactors(self.microdata["growfactors"])
                # apply user specified growdiff
                if self.params["growdiff_response"]:
                    gd_reform.update_growdiff(self.params["growdiff_response"])
                    gd_reform.apply_to(gf_reform)
            reform_records = tc.Records(
                self.microdata["data"],
                start_year=self.microdata["start_year"],
                gfactors=gf_reform,
                weights=self.microdata["weights"],
            )
        else:
            raise ValueError(
                "microdata must be 'CPS', 'PUF', 'TMD', or a dictionary"
            )
        reform_policy = tc.Policy(gf_reform)
        return base_calc, reform_policy, reform_records



"""
setup.py is used to build a light-weight python package around your
Compute Studio code. Add a MANIFEST.in file to specify data files that should
be included with this package. Read more here:
https://docs.python.org/3.8/distutils/sourcedist.html#specifying-the-files-to-distribute
"""

import setuptools
import os

setuptools.setup(
    name="cs-config",
    description="Compute Studio configuration files.",
    url="https://github.com/compute-tooling/compute-studio-kit",
    packages=setuptools.find_packages(),
    include_package_data=True,
)



# Write or import your COMP functions here.
import os
import json
import traceback
import paramtools
import pandas as pd
import taxcalc
import cs2tc
from .constants import MetaParameters
from .helpers import (
    TCDIR,
    postprocess,
    nth_year_results,
    retrieve_puf,
    retrieve_tmd,
)
from .outputs import create_layout, aggregate_plot
from taxbrain import TaxBrain, report
from dask import delayed, compute
from collections import defaultdict, OrderedDict
from marshmallow import fields


AWS_ACCESS_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID")
AWS_SECRET_ACCESS_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY")
PUF_S3_FILE_LOCATION = os.environ.get(
    "PUF_S3_LOCATION", "s3://ospc-data-files/puf.20210720.csv.gz"
)
TMD_S3_FILE_LOCATION = os.environ.get(
    "TMD_S3_LOCATION", "s3://ospc-data-files/tmd.20210720.csv.gz"
)

CUR_PATH = os.path.abspath(os.path.dirname(__file__))


class BehaviorParams(paramtools.Parameters):
    """
    Class for creating behavioral parameters
    """

    array_first = True
    with open(os.path.join(CUR_PATH, "behavior_params.json"), "r") as f:
        behavior_params = json.load(f)
    defaults = behavior_params


def get_version():
    model_versions_str = ""
    for model, version in TaxBrain.VERSIONS.items():
        model_versions_str += f"{model}: {version}\n"
    return model_versions_str


def get_inputs(meta_params_dict):
    """
    Return default parameters for Tax-Brain
    """
    meta_params = MetaParameters()
    with meta_params.transaction(defer_validation=True):
        meta_params.adjust(meta_params_dict)
        # Year must be at least 2014 when using the CPS. This rule is validated
        # in the validate_inputs function below.
        # See: https://github.com/PSLmodels/Tax-Brain/issues/176
        if meta_params.data_source == "CPS" and meta_params.year < 2014:
            meta_params.adjust({"year": 2014})

    policy_params = taxcalc.Policy()
    policy_params.set_state(year=meta_params.year.tolist())

    policy_defaults = cs2tc.convert_policy_defaults(meta_params, policy_params)

    behavior_params = BehaviorParams()

    default_params = {
        "policy": policy_defaults,
        "behavior": behavior_params.dump(),
    }
    meta = meta_params.dump()

    return {"meta_parameters": meta, "model_parameters": default_params}


def validate_inputs(meta_params_dict, adjustment, errors_warnings):
    """
    Function to validate COMP inputs
    """
    meta_params = MetaParameters()
    meta_params.adjust(meta_params_dict, raise_errors=False)
    errors_warnings["policy"]["errors"].update(meta_params.errors)

    pol_params = cs2tc.convert_policy_adjustment(adjustment["policy"])
    policy_params = taxcalc.Policy()
    policy_params.adjust(pol_params, raise_errors=False, ignore_warnings=True)
    errors_warnings["policy"]["errors"].update(policy_params.errors)

    behavior_params = BehaviorParams()
    behavior_params.adjust(adjustment["behavior"], raise_errors=False)
    errors_warnings["behavior"]["errors"].update(behavior_params.errors)

    return {"errors_warnings": errors_warnings}


def run_model(meta_params_dict, adjustment):
    """
    Runs TaxBrain
    """
    # update meta parameters
    meta_params = MetaParameters()
    meta_params.adjust(meta_params_dict)
    # convert COMP user inputs to format accepted by tax-calculator
    policy_mods = cs2tc.convert_policy_adjustment(adjustment["policy"])
    behavior_mods = cs2tc.convert_behavior_adjustment(adjustment["behavior"])
    user_mods = {"policy": policy_mods, "behavior": behavior_mods}
    start_year = int(meta_params.year)
    if meta_params.data_source == "PUF":
        puf_df = retrieve_puf(
            PUF_S3_FILE_LOCATION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY
        )
        if puf_df is not None:
            if not isinstance(puf_df, pd.DataFrame):
                raise TypeError("'puf_df' must be a Pandas DataFrame.")
            fuzz = True
            sampling_frac = 0.05
            sampling_seed = 2222
            full_sample = puf_df
            data_start_year = taxcalc.Records.PUFCSV_YEAR
            weights = taxcalc.Records.PUF_WEIGHTS_FILENAME
        else:
            # Access keys are not available. Default to the CPS.
            print("Defaulting to the CPS")
            meta_params.adjust({"data_source": "CPS"})
    elif meta_params.data_source == "TMD":
        tmd_df = retrieve_tmd(
            TMD_S3_FILE_LOCATION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY
        )
        if tmd_df is not None:
            if not isinstance(tmd_df, pd.DataFrame):
                raise TypeError("'tmd_df' must be a Pandas DataFrame.")
            fuzz = True
            sampling_frac = 0.05
            sampling_seed = 2222
            full_sample = tmd_df
            data_start_year = taxcalc.Records.TMDCSV_YEAR
            weights = taxcalc.Records.TMD_WEIGHTS_FILENAME
        else:
            # Access keys are not available. Default to the CPS.
            print("Defaulting to the CPS")
            meta_params.adjust({"data_source": "CPS"})
    elif meta_params.data_source == "CPS":
        fuzz = False
        input_path = os.path.join(TCDIR, "cps.csv.gz")
        # full_sample = read_egg_csv(cpspath)  # pragma: no cover
        sampling_frac = 0.03
        sampling_seed = 180
        full_sample = pd.read_csv(input_path)
        data_start_year = taxcalc.Records.CPSCSV_YEAR
        weights = os.path.join(taxcalc.Records.CODE_PATH, "cps_weights.csv.gz")
    else:
        raise ValueError(
            f"Data source '{meta_params.data_source}' is not supported."
        )

    if meta_params.use_full_sample:
        sample = full_sample
        end_year = min(start_year + 10, TaxBrain.LAST_BUDGET_YEAR)
    else:
        sample = full_sample.sample(
            frac=sampling_frac, random_state=sampling_seed
        )
        end_year = start_year

    tb = TaxBrain(
        start_year,
        end_year,
        microdata={
            "data": sample,
            "start_year": data_start_year,
            "growfactors": None,
            "weights": weights,
        },
        reform=policy_mods,
        behavior=behavior_mods,
    )
    tb.run()

    # Collect results for each year
    delayed_list = []
    for year in range(start_year, end_year + 1):
        print("delaying for", year)
        delay = delayed(nth_year_results)(tb, year, user_mods, fuzz)
        delayed_list.append(delay)
    results = compute(*delayed_list)

    # process results to get them ready for display
    # create aggregate plot
    agg_plot = aggregate_plot(tb)
    all_to_process = defaultdict(list)
    for result in results:
        for key, value in result.items():
            all_to_process[key] += value
    results, downloadable = postprocess(all_to_process)
    # create report output if it is not a run with no reforme
    if tb.params["policy"].keys():
        report_outputs = report(tb, clean=True)
        for name, data in report_outputs.items():
            if name.endswith(".md"):
                media_type = "Markdown"
            elif name.endswith(".pdf"):
                media_type = "PDF"
            downloadable.append(
                {"media_type": media_type, "title": name, "data": data}
            )
    agg_output, table_output = create_layout(results, start_year, end_year)

    comp_outputs = {
        "renderable": [agg_plot, agg_output, table_output],
        "downloadable": downloadable,
    }
    return comp_outputs



from typing import Union, Mapping, Any, List, Tuple

import paramtools


TaxcalcReform = Union[str, Mapping[int, Any]]
ParamToolsAdjustment = Union[str, List[paramtools.ValueObject]]
PlotColors = Union[str, Tuple[float, float, float]]
